from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.ev_user import EvUser, EvUserL, EvUserD


class TestEagerValidation(TestCase):

    def test_eager_modifier_validates_on_init(self):
        with self.assertRaises(ValidationException) as context:
            EvUser(password='123')
        self.assertEqual("length of value is not greater than or equal 8",
                         context.exception.keypath_messages['password'])

    def test_eager_modifier_doesnt_cause_other_fields_to_validate_on_init(self):
        user = EvUser(password='12345678')
        self.assertEqual(user._data_dict, {'username': None, 'password': '123456780x0x'})

    def test_eager_modifier_will_not_perform_when_value_is_none_on_init(self):
        try:
            EvUser()
        except ValidationException:
            self.fail('eager modifier should not perform on init if value is None.')

    def test_eager_modifier_validates_on_set(self):
        user = EvUser()
        with self.assertRaises(ValidationException) as context:
            user.set(password='123')
        self.assertEqual("length of value is not greater than or equal 8",
                         context.exception.keypath_messages['password'])

    def test_eager_modifier_will_not_work_on_update(self):
        user = EvUser()
        try:
            user.update(password='123')
        except ValidationException:
            self.fail('eager modifier should not perform on update.')

    def test_eager_modifier_prevents_modifiers_before_it_on_validate(self):
        user = EvUser(password='1234567890123456', username='Tsi')
        try:
            user.validate()
        except ValidationException:
            self.fail('eager modifier should prevent modifiers before it to work on validate.')

    def test_eager_modifier_should_work_inside_list(self):
        try:
            EvUserL(passwords=['123', '456', '789', '012'])
        except ValidationException:
            self.fail('eager modifier should not throw if value is valid')

    def test_eager_modifier_should_throw_inside_list(self):
        with self.assertRaises(ValidationException):
            EvUserL(passwords=['123xxx', '456xxx', '789xxx', '012xxx'])

    def test_eager_modifier_should_work_inside_dict(self):
        try:
            EvUserD(passwords={'a': '123', 'b': '456', 'c': '789', 'd': '012'})
        except ValidationException:
            self.fail('eager modifier should not throw if value is valid')

    def test_eager_modifier_should_validate_and_throw_inside_dict(self):
        with self.assertRaises(ValidationException):
            EvUserD(passwords={'a': '123xxx', 'b': '456xxx', 'c': '789xxx',
                               'd': '012xxx'})

    def test_eager_modifier_should_lazy_validate_when_validate_inside_list(self):
        user = EvUserL(passwords=['123', '456', '789', '012'])
        try:
            user.validate()
        except ValidationException:
            self.fail('eager modifier should not throw if not validation task after eager mark')

    def test_eager_modifier_should_lazy_validate_when_validate_inside_dict(self):
        user = EvUserD(passwords={'a': '123', 'b': '456', 'c': '789', 'd': '012'})
        try:
            user.validate()
        except ValidationException:
            self.fail('eager modifier should not throw if not validation task after eager mark')



from __future__ import annotations
from jsonclasses import jsonclass, types


__id = 0


def nextid():
    global __id
    __id = __id + 1
    return __id


@jsonclass
class LKAuthor:
    id: int = types.int.readonly.primary.default(nextid).required
    name: str
    articles: list[LKArticle] = types.listof('LKArticle').linkedby('author')


@jsonclass
class LKArticle:
    id: int = types.int.readonly.primary.default(nextid).required
    name: str
    author: LKAuthor = types.objof(LKAuthor).linkto



"""module for tobosec modifier."""
from __future__ import annotations
from datetime import datetime
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToBoSecModifier(Modifier):
    """
    ToBoSec Modifier transforms datetime into the beginning of the second
    """

    def transform(self, ctx: Ctx) -> Any:
        return ctx.val.replace(microsecond=0) if type(ctx.val) is datetime else ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.calc_user import CalcUser


class TestGetter(TestCase):

    def test_getter_gets_correct_value(self):
        user = CalcUser(name="Peter Layber", base_score=25.54)
        self.assertEqual(user.first_name, 'Peter')
        self.assertEqual(user.last_name, 'Layber')
        self.assertAlmostEqual(user.score, 51.08)
        self.assertEqual(user._data_dict,
                         {'name': 'Peter Layber', 'base_score': 25.54})

    def test_calc_fields_can_be_output_to_json(self):
        user = CalcUser(name="Peter Layber", base_score=25.54)
        result = user.tojson()
        self.assertEqual(result, {'name': 'Peter Layber',
                                  'firstName': 'Peter',
                                  'lastName': 'Layber',
                                  'score': 51.08,
                                  'baseScore': 25.54})

    def test_calc_fields_can_be_validated(self):
        user = CalcUser(name="Peter Layber", base_score=25.54)
        with self.assertRaises(ValidationException) as context:
            user.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['score'],
                         "value is not negative")



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_round import SuperRound


class TestFloor(TestCase):

    def test_floor_transforms_float_value_into_int_by_flooring(self):
        sr = SuperRound(floor_value=19.8)
        self.assertEqual(sr.floor_value, 19)

    def test_floor_keeps_int_value_unchanged(self):
        sr = SuperRound(floor_value=40)
        self.assertEqual(sr.floor_value, 40)



from __future__ import annotations
from tests.classes.super_iterable import SuperIterable
from unittest import TestCase


class TestJoin(TestCase):

    def test_join_keeps_str_value_if_it_is_str(self):
        s = SuperIterable(itsj='asdhakjsh')
        self.assertEqual(s.itsj, 'asdhakjsh')

    def test_join_joins_a_list_of_str_into_a_str(self):
        s = SuperIterable(itsj=["a", "d", "s", "r"])
        self.assertEqual(s.itsj, 'a-d-s-r')

    def test_Join_keeps_callable_value_if_it_is_callable(self):
        s = SuperIterable(c_itsj='asdhakjsh')
        self.assertEqual(s.c_itsj, 'asdhakjsh')

    def test_join_joins_a_types_into_a_types(self):
        s = SuperIterable(t_itsj=["a", "d", "s", "r"])
        self.assertEqual(s.t_itsj, 'a-d-s-r')



"""module for random digits modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from random import sample
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class RandomDigitsModifier(Modifier):
    """Random digits modifier generates random digits as result."""

    def __init__(self, length: int | Callable | Types) -> None:
        self.length = length

    def transform(self, ctx: Ctx) -> Any:
        digits = '0123456789'
        return ''.join(sample(digits, self.resolve_param(self.length, ctx)))



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.email_user import EmailUser


class TestEmail(TestCase):

    def test_simple_email_doesnt_raise(self):
        analysis = EmailUser(username='hello', email='hello@qq.com')
        analysis.validate()

    def test_email_with_underbar_doesnt_raise(self):
        analysis = EmailUser(username='hello', email='hello_World@qq.com.cn')
        analysis.validate()

    def test_email_with_dot_doesnt_raise(self):
        analysis = EmailUser(username='hello', email='hello.world@qq.com.cn')
        analysis.validate()

    def test_email_with_single_letter_in_each_field_doesnt_raise(self):
        analysis = EmailUser(username='hello', email='a@a.a.cn')
        analysis.validate()

    def test_email_with_multiple_domains_doesnt_raise(self):
        analysis = EmailUser(username='hello', email='a@a.a.cn.a.a')
        analysis.validate()

    def test_email_raises_if_value_is_not_valid_email(self):
        analysis = EmailUser(username='hello', email='@qq.com')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['email'],
                         "value is not email string")

    def test_email_raises_if_email_contains_special_charaters(self):
        analysis = EmailUser(username='hello', email='!a@qq.com')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['email'],
                         "value is not email string")

    def test_email_raises_exception_if_email_contains_hashtag(self):
        analysis = EmailUser(username='hello', email='#@qq.com')
        with self.assertRaises(ValidationException) as context:
            analysis.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['email'],
                         "value is not email string")



from __future__ import annotations
from typing import Any
from os import getcwd
from uuid import uuid4
from pathlib import Path
from .uconf import uconf, UserConf


_uploaders: dict[str, Uploader] = {}


class Uploader:

    def __init__(self, config: dict[str, Any]) -> None:
        self.config = config

    def upload(self, value: Any) -> str:
        raise Exception('please use concrete subclass of uploader')


class S3Uploader(Uploader):

    @property
    def client(self) -> Any:
        from boto3 import client as s3_client
        if hasattr(self, '_client'):
            return getattr(self, '_client')
        client = s3_client('s3',
                region_name=self.config['region_name'],
                endpoint_url=self.config['endpoint_url'],
                aws_access_key_id=self.config['aws_access_key_id'],
                aws_secret_access_key=self.config['aws_secret_access_key'])
        setattr(self, '_client', client)
        return client

    def upload(self, value: Any) -> str:
        unique_filename = f'{uuid4().hex}{Path(value.filename).suffix.lower()}'
        if value.__class__.__name__ == 'UploadFile':
            uploadobj = value.file
            content_type = value.content_type
        else:
            uploadobj = value
            content_type = value.content_type
        self.client.upload_fileobj(uploadobj, self.config['bucket'], unique_filename, ExtraArgs={
            'ACL': self.config['acl'],
            'ContentType': content_type
        })
        return f"{self.config['endpoint_url']}/{self.config['bucket']}/{unique_filename}"


class AliOSSUploader(Uploader):

    def upload(self, value: Any) -> str:
        pass


class LocalFSClient:

    def __init__(self, dir: str, url: str) -> None:
        self.dir = dir
        self.url = url
        dest = Path(getcwd()) / 'public' / self.dir
        if not dest.is_dir():
            dest.mkdir(parents=True)
        self.dest = dest

    def upload(self, value: Any) -> str:
        unique_filename = f'{uuid4().hex}{Path(value.filename).suffix.lower()}'
        if value.__class__.__name__ == 'UploadFile':
            file_location = str(self.dest / unique_filename)
            with open(file_location, "wb+") as file_object:
                file_object.write(value.file.read())
        else:
            value.save(str(self.dest / unique_filename))
        return self.url + '/public/' + self.dir + '/' + unique_filename


class LocalFSUploader(Uploader):

    @property
    def client(self) -> Any:
        if hasattr(self, '_client'):
            return getattr(self, '_client')
        client = LocalFSClient(dir=self.config['dir'], url=self.config['url'])
        setattr(self, '_client', client)
        return client

    def upload(self, value: Any) -> str:
        return self.client.upload(value)


def request_uploader(name: str) -> Uploader:
    if _uploaders.get(name):
        return _uploaders[name]
    conf = uconf()
    uploaders_dict = conf.get('uploaders')
    if uploaders_dict is None:
        raise ValueError(f"uploader '{name}' is undefined")
    uploader_conf = uploaders_dict.get(name)
    if uploader_conf is None:
        raise ValueError(f"uploader '{name}' is undefined")
    engine = uploader_conf.get('client')
    if engine is None:
        raise ValueError(f"undefined client engine for '{name}'")
    config = uploader_conf.get('config')
    if config is None:
        raise ValueError(f"config is not found for uploader '{name}'")
    match engine:
        case 's3':
            uploader = S3Uploader(config)
            _uploaders[name] = uploader
            return uploader
        case 'alioss':
            uploader = AliOSSUploader(config)
            _uploaders[name] = uploader
            return uploader
        case 'localfs':
            uploader = LocalFSUploader(config)
            _uploaders[name] = uploader
            return uploader
        case _:
            raise ValueError(f"unexist uploader engine '{engine}'")



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class SuperMap:
    list1: list[int] | None = types.listof(int).map(lambda i: i + 1)
    list2: list[int] | None = types.listof(int).map(types.add(1))



"""module for preserialize modifier."""
from .modifier import Modifier
from ..fdef import FDef


class PreserializeModifier(Modifier):
    """A PreserializeModifier tweaks field validation logic. Every modifier
    after a preserialize modifier are only triggered just before serialization
    into database.

    This is usually used before setonsave modifier.
    """

    def define(self, fdef: FDef) -> None:
        fdef._has_preserialize_modifier = True



from __future__ import annotations
from random import random
from tests.classes.super_random import SuperRandom
from unittest import TestCase

class TestRandomDigits(TestCase):

    def test_random_digits_generates_random_digits_as_result(self):
        i = SuperRandom(random_digits=1)
        self.assertEqual(type(i.random_digits), type('1'))

    def test_random_digits_generates_random_callable_as_result(self):
        i = SuperRandom(c_random_digits=1)
        self.assertEqual(type(i.c_random_digits), type('1'))

    def test_random_digits_generates_random_types_as_result(self):
        i = SuperRandom(t_random_digits=1)
        self.assertEqual(type(i.t_random_digits), type('1'))



"""module for tofloat modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ToFloatModifier(Modifier):
    """ToFloat Modifier transforms value into a float"""

    def transform(self, ctx: Ctx) -> Any:
        is_str_or_bool = type(ctx.val) is bool or type(ctx.val) is str
        return float(ctx.val) if is_str_or_bool else ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from classes.zero_number import ZeroNumber


class TestPositive(TestCase):
    def test_positive_doesnt_raise_if_float_value_is_greater_than_zero(self):
        n = ZeroNumber(fpositive=5.5)
        n.validate()

    def test_positive_raises_if_float_value_is_equal_to_zero(self):
        n = ZeroNumber(fpositive=0)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['fpositive'],
                         "value is not positive")

    def test_positive_raises_if_float_value_is_less_than_zero(self):
        n = ZeroNumber(fpositive=-5.5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['fpositive'],
                         "value is not positive")

    def test_positive_doesnt_raise_if_int_value_is_greater_than_zero(self):
        n = ZeroNumber(ipositive=5)
        n.validate()

    def test_positive_raises_if_int_value_is_equal_to_zero(self):
        n = ZeroNumber(ipositive=0)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ipositive'],
                         "value is not positive")

    def test_positive_raises_if_int_value_is_less_than_zero(self):
        n = ZeroNumber(ipositive=-5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ipositive'],
                         "value is not positive")



from __future__ import annotations
from tests.classes.super_str import SuperStr
from unittest import TestCase



class TestPadStart(TestCase):

    def test_padstart_adds_str_to_the_start_of_str(self):
        ps = SuperStr(pads="aaaaa")
        self.assertEqual(ps.pads, "sssssaaaaa")

    def test_padstart_keeps_value_when_type_of_value_is_not_str(self):
        ps = SuperStr(padis=80)
        self.assertEqual(ps.padis, 80)

    def test_padstart_keeps_value_when_target_length_less_value_length(self):
        ps = SuperStr(pads="aaaaaaaaaaa")
        self.assertEqual(ps.pads, "aaaaaaaaaaa")

    def test_padstart_adds_callable_to_the_start_of_str(self):
        ps = SuperStr(c_pads="aaaaa")
        self.assertEqual(ps.c_pads, "sssssaaaaa")

    def test_padstart_adds_types_to_the_start_of_str(self):
        ps = SuperStr(t_pads="aaaaa")
        self.assertEqual(ps.t_pads, "sssssaaaaa")



from __future__ import annotations
from unittest import TestCase
from tests.classes.cb_product import CBProduct
from tests.classes.cbm_product import CBMProduct
from tests.classes.cbo_product import CBOProduct


class TestDecoratedOnDelete(TestCase):

    def test_callback_are_called_for_existing_objects_on_delete(self):
        p = CBProduct(name='N')
        setattr(p, '_is_new', False)
        p.delete()
        self.assertEqual(p.deleted_count, 99)

    def test_multiple_callbacks_are_called_for_existing_objects_on_del(self):
        p = CBMProduct(name='N')
        setattr(p, '_is_new', False)
        p.delete()
        self.assertEqual(p.deleted_count, 98)

    def test_operator_can_be_passed_into_callback(self):
        p = CBOProduct(name='N')
        setattr(p, '_is_new', False)
        p.opby(10)
        p.delete()
        self.assertEqual(p.deleted_count, 90)



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_type import SuperType


class TestToBool(TestCase):

    def test_tobool_returns_true_if_value_evaluates_to_true(self):
        b = SuperType(tb={1, 3, 4})
        self.assertEqual(b.tb, True)

    def test_tobool_returns_false_if_value_evaluates_to_false(self):
        b = SuperType(tb="")
        self.assertEqual(b.tb, False)



from __future__ import annotations
from datetime import datetime
from unittest import TestCase
from tests.classes.super_datetime import SuperDateTime

class TestTobosec(TestCase):

    def test_tobosec_transforms_datetime_into_the_time_of_beginning_of_second(self):
        d = SuperDateTime(dtbs=datetime(2021, 10, 11, 17, 37, 27, 446259))
        self.assertEqual(d.dtbs, datetime(2021,10, 11, 17, 37, 27))

    def test_tobosec_does_not_transform_if_is_not_datetime(self):
        s = SuperDateTime(stbs="12345")
        self.assertEqual(s.stbs, "12345")



from __future__ import annotations
from unittest import TestCase
from tests.classes.user_onsave import (UserOnsave, UserOnsaveZ, check_value,
                                       check_args)


class TestOnSave(TestCase):

    def test_onsave_is_triggered_for_new_objects(self):
        val = check_value()
        user = UserOnsave(name='Lao', age=6)
        user.name = 'Ua Tsiu Liong Bo Ho'
        user._set_on_save()
        self.assertEqual(val + 1, check_value())
        self.assertEqual(check_args(), 6)

    def test_onsave_is_triggered_for_modified_objects(self):
        val = check_value()
        user = UserOnsave(name='Lao', age=7)
        setattr(user, '_is_new', False)
        user.name = 'Ua Tsiu Liong Bo Ho'
        user._set_on_save()
        self.assertEqual(val + 1, check_value())
        self.assertEqual(check_args(), 7)

    def test_onsave_is_not_triggered_for_unmodified_objects(self):
        val = check_value()
        user = UserOnsave(name='Lao', age=5)
        setattr(user, '_is_new', False)
        user._set_on_save()
        self.assertEqual(val, check_value())
        self.assertNotEqual(check_args(), 5)

    def test_onsave_can_take_no_args(self):
        val = check_value()
        user = UserOnsaveZ(name='Cong Ti', age=10)
        setattr(user, '_is_new', False)
        user.name = 'Mai Kah Ua Cong Kha'
        user._set_on_save()
        self.assertEqual(val + 10, check_value())
        self.assertNotEqual(check_args(), 10)



# pylint: skip-file
"""JSON Class Mypy plugin for type checking JSON Class classes and objects."""
from typing import Optional, Callable, Any
from mypy.plugin import Plugin, ClassDefContext
from mypy.options import Options
from mypy.types import AnyType, TypeOfAny
from mypy.nodes import (CallExpr, LambdaExpr, MemberExpr, TempNode, TypeInfo,
                        AssignmentStmt, NameExpr, PlaceholderNode, Var)
from mypy.errorcodes import ErrorCode

JSONCLASS_DECORATOR_FULLNAME = 'jsonclasses.jsonclass.jsonclass'
JSONCLASS_TYPES_FULLNAME = 'jsonclasses.types.types'

ERROR_UNTYPED = ErrorCode('jsonclass-field', 'Untyped field disallowed', 'JSON Class')
ERROR_MULTIPLE_DEFAULT_VALUES = ErrorCode('jsonclass-field',
                                          'Multiple default values defined',
                                          'JSON Class')


def is_json_class_types_expr(expr: Any) -> bool:
    if isinstance(expr, NameExpr):
        return expr.fullname == JSONCLASS_TYPES_FULLNAME
    if isinstance(expr, MemberExpr):
        return is_json_class_types_expr(expr.expr)
    if isinstance(expr, CallExpr):
        return is_json_class_types_expr(expr.callee)
    return False


def json_class_types_default_arg_expr(expr: Any) -> Any:
    if isinstance(expr, NameExpr):
        return None
    if isinstance(expr, MemberExpr):
        return json_class_types_default_arg_expr(expr.expr)
    if isinstance(expr, CallExpr):
        if isinstance(expr.callee, MemberExpr) and expr.callee.name == 'default':
            return expr.args
        else:
            return json_class_types_default_arg_expr(expr.callee)
    return None


def collect_attrs(ctx: ClassDefContext):
    pass


def transform_jsonclass_class(ctx: ClassDefContext) -> None:
    cls = ctx.cls
    for stmt in cls.defs.body:
        if not isinstance(stmt, AssignmentStmt):
            continue
        lhs = stmt.lvalues[0]
        if not isinstance(lhs, NameExpr):
            continue
        if not stmt.new_syntax:
            ctx.api.fail('Untyped fields disallowed in JSON Class body.', stmt, code=ERROR_UNTYPED)
            continue
        sym = cls.info.names.get(lhs.name)
        if sym is None:
            # This name is likely blocked by a star import. We don't need to defer because
            # defer() is already called by mark_incomplete().
            continue
        node = sym.node
        if isinstance(node, PlaceholderNode):
            # This node is not ready yet.
            return None
        assert isinstance(node, Var)
        if node.is_classvar:
            continue
        # node_type = get_proper_type(node.type)
        if is_json_class_types_expr(stmt.rvalue):
            stmt.rvalue = TempNode(AnyType(TypeOfAny.explicit), False)
            # default_expr = json_class_types_default_arg_expr(stmt.rvalue)
            # if default_expr:
            #     if len(default_expr) > 1:
            #         ctx.api.fail('Multiple default values defined.',
            #                      default_expr[1], code=ERROR_MULTIPLE_DEFAULT_VALUES)
            #         stmt.rvalue = TempNode(AnyType(TypeOfAny.explicit), False)
            #         continue
            #     expr = default_expr[0]
            #     if not (isinstance(expr, LambdaExpr) or isinstance(expr, MemberExpr)):
            #         # TODO: check callable return type
            #         stmt.rvalue = expr
            #     else:
            #         stmt.rvalue = TempNode(AnyType(TypeOfAny.explicit), False)
            # else:
            #     stmt.rvalue = TempNode(AnyType(TypeOfAny.explicit), False)


class JSONClassesPlugin(Plugin):

    def __init__(self, options: Options) -> None:
        super().__init__(options)

    def get_class_decorator_hook(self, fullname: str) -> Optional[Callable[[ClassDefContext], None]]:
        if fullname == JSONCLASS_DECORATOR_FULLNAME:
            return transform_jsonclass_class
        else:
            return super().get_class_decorator_hook(fullname)


def plugin(version: str) -> type[Plugin]:
    """Returns the JSON Class Mypy plugin. The version argument is ignored.
    """
    return JSONClassesPlugin



from __future__ import annotations
from jsonclasses import jsonclass


def set_deleted(p: CBProduct) -> None:
    p.deleted_count = p.deleted_count - 1


@jsonclass(on_delete=set_deleted)
class CBProduct:
    name: str
    deleted_count: int = 100



"""module for alpha modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class AlphaModifier(Modifier):
    """Alpha modifier raises if value is not alpha string."""

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, str) and not ctx.val.isalpha():
            ctx.raise_vexc('value is not alpha str')



from __future__ import annotations
from jsonclasses import jsonclass, types


def check_author(book: GSBook, operator: GSBookAuthor) -> bool:
    return book.author.id == operator.id and operator.paid_user


@jsonclass
class GSBookAuthor:
    id: str
    name: str
    paid_user: bool
    books: list[GSBook] = types.nonnull.listof('GSBook').linkedby('author')


@jsonclass(can_update=check_author)
class GSBook:
    name: str
    author: GSBookAuthor = types.objof('GSBookAuthor').linkto.required


@jsonclass(can_update=types.getop.isobjof('GSBookAuthor'))
class GSTBook:
    name: str



"""module for canr modifier."""
from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
from ..fdef import FDef
from ..excs import ValidationException
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class CanRModifier(Modifier):
    """Whether this operator can read on this field.
    """

    def __init__(self, checker: Callable | Types) -> None:
        self.checker = checker

    def tojson(self, ctx: Ctx) -> Any:
        super().tojson(ctx)
        if ctx.operator is None:
            return None
        if callable(self.checker):
            params_len = len(signature(self.op_callable).parameters)
            if params_len == 1:
                result = self.op_callable(ctx.operator)
            elif params_len == 2:
                result = self.op_callable(ctx.operator, ctx.owner)
            elif params_len == 3:
                result = self.op_callable(ctx.operator, ctx.owner, ctx.val)
            elif params_len == 4:
                result = self.op_callable(ctx.operator, ctx.owner, ctx.val, ctx)
            if result is None:
                return ctx.val
            if result is True:
                return ctx.val
            if result is False:
                return None
            if isinstance(result, str):
                return None
            raise ValueError('invalid modifier')
        else:
            transformed = self.checker.modifier.transform(ctx)
            try:
                self.checker.modifier.validate(ctx.nval(transformed))
                return ctx.val
            except ValidationException:
                return None



"""module for readwrite modifier."""
from ..fdef import FDef, WriteRule, ReadRule
from .modifier import Modifier


class ReadwriteModifier(Modifier):
    """Readwrite modifier marks a field both readable and writable."""

    def define(self, fdef: FDef) -> None:
        fdef._write_rule = WriteRule.UNLIMITED
        fdef._read_rule = ReadRule.UNLIMITED



"""This module defineds the JSON class mapping graph."""
from __future__ import annotations
from typing import final, TYPE_CHECKING
from enum import Enum
from .jconf import JConf
from .keypath import camelize_key, reference_key, underscore_key
from .excs import (JSONClassRedefinitionException,
                         JSONClassTypedDictRedefinitionException,
                         JSONClassNotFoundException,
                         JSONClassTypedDictNotFoundException)
if TYPE_CHECKING:
    from .cdef import CDef


@final
class CGraph:
    """JSON classes are defined on class graphs. Classes in the same graph
    share default configurations and can interoperate with each other. A class
    is guaranteed to have a unique name in it's graph. Different graphs serve
    as different naming spaces.
    """

    _graph_map: dict[str, CGraph] = {}
    """The graph map on which graph objects are stored."""

    _initialized_map: dict[str, bool] = {}
    """The map on which graphs' initialization status are recorded."""

    def __new__(cls: type[CGraph], name: str) -> CGraph:
        """The `CGraph` class returns a shared graph object by it's
        name.
        """
        if not cls._graph_map.get(name):
            cls._graph_map[name] = super(CGraph, cls).__new__(cls)
        return cls._graph_map.get(name)

    def __init__(self: CGraph, name: str) -> None:
        """Find a class graph by it's name. A new one is created if it's not
        exist.

        Args:
            name (str): The name of the graph.
        """
        if self.__class__._initialized_map.get(name):
            return
        self._name: str = name
        self._map: dict[str, CDef] = {}
        self._enum_map: dict[str, type] = {}
        self._default_config = JConf(cgraph=self.name,
                                     output_key_strategy=camelize_key,
                                     input_key_strategy=underscore_key,
                                     strict_input=True,
                                     ref_name_strategy=reference_key,
                                     validate_all_fields=False,
                                     abstract=False,
                                     reset_all_fields=False,
                                     output_null=False,
                                     on_create=[],
                                     on_update=[],
                                     on_delete=[],
                                     can_create=[],
                                     can_update=[],
                                     can_delete=[],
                                     can_read=[])
        self.__class__._initialized_map[name] = True
        return None

    def __repr__(self) -> str:
        return f'[{self._name}]'

    @property
    def name(self: CGraph) -> str:
        """The name of this class graph."""
        return self._name

    @property
    def default_config(self: CGraph) -> JConf:
        """The default configuration used on this class graph."""
        return self._default_config

    def put(self: CGraph, cdef: CDef) -> None:
        """Put a class onto this class graph.

        Args:
            class_ (type): The JSON class which will be put onto this graph.

        Raises:
            JSONClassRedefinitionException: This exception is raised if a \
                new class with existing name is defined.
        """
        exist_definition = self._map.get(cdef.name)
        if exist_definition:
            raise JSONClassRedefinitionException(cdef.cls,
                                                 exist_definition.cls)
        self._map[cdef.name] = cdef

    def fetch(self: CGraph, name_or_class: str | type) -> CDef:
        """Fetch a class by it's name from this class graph.

        Args:
            name_or_class (Union[str, type]): The name of the class to be \
                fetched or the class itself.

        Raises:
            JSONClassNotFoundException: This exception is raised if a class \
                definition with `name` is not found.
        """
        if isinstance(name_or_class, type):
            name = name_or_class.__name__
        else:
            name = name_or_class
        try:
            return self._map[name]
        except KeyError:
            raise JSONClassNotFoundException(name, self.name)

    def has(self: CGraph, name_or_class: str | type) -> bool:
        """Test if class with name is registered in the graph.

        Args:
            name_or_class (Union[str, type]): The name of the class to be \
                fetched or the class itself.

        Returns:
            bool: Returns True if this class is registered in the graph.
        """
        if isinstance(name_or_class, type):
            name = name_or_class.__name__
        else:
            name = name_or_class
        return self._map.get(name) is not None

    def put_enum(self: CGraph, enum_class: type[Enum]) -> None:
        """Put a enum class onto this class graph.

        Args:
            enum_class (type): The enum class which will be put onto this
            graph.

        Raises:
            JSONClassRedefinitionException: This exception is raised if a \
                new enum class with existing name is defined.
        """
        exist_def = self._enum_map.get(enum_class.__name__)
        if exist_def:
            raise JSONClassTypedDictRedefinitionException(enum_class,
                                                          exist_def.cls)
        self._enum_map[enum_class.__name__] = enum_class

    def fetch_enum(self: CGraph, ec_or_name: type[Enum] | str) -> type[Enum]:
        """Fetch a enum class by it's name from this class graph.

        Args:
            ec_or_name (Union[type[Enum], str]): The name of the enum class to
            be fetched or the class itself.

        Raises:
            JSONClassNotFoundException: This exception is raised if a class \
                definition with `name` is not found.
        """
        if isinstance(ec_or_name, type):
            name = ec_or_name.__name__
        else:
            name = ec_or_name
        try:
            return self._enum_map[name]
        except KeyError:
            raise JSONClassTypedDictNotFoundException(name, self.name)

    def has_enum(self: CGraph, ec_or_name: type[Enum] | str) -> bool:
        """Test if a enum class with name is registered in the graph.

        Args:
            ec_or_name (Union[type[dict], str]): The name of the enum class to
            be fetched or the class itself.

        Returns:
            bool: Returns True if this enum class is registered in the graph.
        """
        if isinstance(ec_or_name, type):
            name = ec_or_name.__name__
        else:
            name = ec_or_name
        return self._enum_map.get(name) is not None



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperAppend:
    s: Optional[str] = types.str.append('3432')
    l: Optional[list[str]] = types.listof(str).append('7788')

    cs: Optional[str] = types.str.append(lambda: '3432')
    cl: Optional[list[str]] = types.listof(str).append(lambda: '7788')

    ts: Optional[str] = types.str.append(types.default('3432'))
    tl: Optional[list[str]] = types.listof(str).append(types.default('7788'))



"""module for datetime modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from datetime import date, datetime, timezone
from ..fdef import FType
from .type_modifier import TypeModifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class DatetimeModifier(TypeModifier):
    """Datetime modifier validate value against datetime type."""

    def __init__(self):
        super().__init__()
        self.cls = datetime
        self.ftype = FType.DATETIME

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        elif isinstance(ctx.val, str):
            try:
                retval = datetime.fromisoformat(ctx.val.replace('Z', ''))
                return retval.replace(tzinfo=timezone.utc)
            except ValueError:
                ctx.raise_vexc('wrong datetime format')
        elif type(ctx.val) is date:
            retval = datetime(ctx.val.year,
                              ctx.val.month,
                              ctx.val.day, 0, 0, 0)
            return retval.replace(tzinfo=timezone.utc)
        else:
            return ctx.val

    def tojson(self, ctx: Ctx) -> Any:
        return None if ctx.val is None else ctx.val.isoformat()[:23] + 'Z'



from __future__ import annotations
from typing import Optional
from datetime import datetime
from jsonclasses import jsonclass, types


@jsonclass(class_graph='blog')
class User:
    id: int = types.int.primary.required
    name: str
    posts: list[Post] = types.nonnull.listof('Post').linkedby('user').required
    comments: list[Comment] = types.listof('Comment').linkedby('commenter') \
                                   .required
    updated_at: datetime = types.datetime.tsupdated.required


@jsonclass(class_graph='blog')
class Post:
    id: int = types.int.primary.required
    name: str
    user: User = types.linkto.objof('User').required
    comments: list[Comment] = types.listof('Comment').linkedby('post').required
    updated_at: datetime = types.datetime.tsupdated.required



@jsonclass(class_graph='blog')
class Comment:
    id: int = types.int.primary.required
    content: str
    post: Post = types.linkto.objof('Post').required
    parent: Optional[Comment] = types.linkto.objof('Comment')
    children: list[Comment] = types.listof('Comment').linkedby('parent') \
                                   .required
    commenter: User = types.linkto.objof('User').required
    updated_at: datetime = types.datetime.tsupdated.required



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.valid_password import (
    ValidPassword, ValidPasswordMessage, CValidPassword, OptionalPassword,
    TValidPassword
)


class TestValidateModifier(TestCase):

    def test_validate_wont_validate_none(self):
        opw = OptionalPassword(name='Bo Lang', password=None)
        opw.validate()

    def test_validate_is_fine_when_modifier_returns_true(self):
        pw = ValidPassword(name='Li Si', password='0000')
        try:
            pw.validate()
        except ValidationException:
            self.fail('validate should be fine if value is valid')

    def test_validate_raises_default_msg_when_modifier_returns_false(self):
        pw = ValidPassword(name='Li Si', password='000')
        self.assertRaisesRegex(ValidationException,
                               'invalid value',
                               pw.validate)

    def test_validate_raises_if_modifier_returns_str(self):
        pw = ValidPasswordMessage(name='Li Si', password='000')
        self.assertRaisesRegex(ValidationException, 'wrong', pw.validate)

    def test_validate_is_fine_when_modifier_returns_none(self):
        pw = ValidPasswordMessage(name='Li Si', password='0000')
        pw.validate()

    def test_validate_can_also_accept_context(self):
        pw = CValidPassword(name='Li Si', password='0000')
        pw.validate()

    def test_validate_doesnt_raise_if_types_validator_is_valid(self):
        pw = TValidPassword(name='Q', password=45)
        pw.validate()

    def test_validate_raises_if_types_validator_isnt_valid(self):
        pw = TValidPassword(name='Q', password=50)
        self.assertRaisesRegex(ValidationException, 'invalid value', pw.validate)



from __future__ import annotations
from typing import Optional
from datetime import date
from jsonclasses import jsonclass


@jsonclass
class SimpleDeadline:
    ended_at: Optional[date]
    message: Optional[str]



from __future__ import annotations
from datetime import datetime
from unittest import TestCase
from tests.classes.super_datetime import SuperDateTime

class TestTonextsec(TestCase):

    def test_tonextsec_transforms_datetime_into_the_time_of_next_second(self):
        d = SuperDateTime(dtns=datetime(2021, 10, 11, 17, 37, 27, 446259))
        self.assertEqual(d.dtns, datetime(2021,10, 11, 17, 37, 28))

    def test_tonextsec_does_not_transform_if_is_not_datetime(self):
        s = SuperDateTime(stns="12345")
        self.assertEqual(s.stns, "12345")



"""module for listof modifier."""
from __future__ import annotations
from typing import Any, Collection, Iterable, TYPE_CHECKING
from ..fdef import FDef, FStore, FType, Nullability
from .collection_type_modifier import CollectionTypeModifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ListOfModifier(CollectionTypeModifier):
    """This modifier validates list."""

    def __init__(self, raw_item_types: Any) -> None:
        super().__init__(raw_item_types)
        self.cls = list
        self.ftype = FType.LIST

    def define(self, fdef: FDef) -> None:
        super().define(fdef)
        if fdef._fstore == FStore.LOCAL_KEY or fdef._foreign_key is not None:
            if fdef._collection_nullability is Nullability.UNDEFINED:
                fdef._collection_nullability = Nullability.NONNULL

    def enumerator(self, value: list) -> Iterable:
        return enumerate(value)

    def empty_collection(self) -> Collection:
        return []

    def append_value(self, i: int, v: Any, col: list):
        col.append(v)

    def should_special_handle(self, key: Any, v: Any, ctx: Ctx) -> bool:
        is_lkey = ctx.fdef.fstore == FStore.LOCAL_KEY
        is_fkey = ctx.fdef.fstore == FStore.FOREIGN_KEY
        uses_jt = ctx.fdef.use_join_table
        if not ((is_fkey and uses_jt) or is_lkey):
            return False
        return isinstance(v, dict) and ('_add' in v or '_del' in v)

    def special_handle(self, key: Any, v: Any, ctx: Ctx) -> None:
        fname = ctx.keypatho[-1]
        is_lkey = ctx.fdef.fstore == FStore.LOCAL_KEY
        if is_lkey:
            if '_add' in v:
                ctx.owner._link_local_keys(fname, v['_add'])
            elif '_del' in v:
                ctx.owner._unlink_local_keys(fname, v['_del'])
        else:
            if '_add' in v:
                ctx.owner._add_link_key(fname, v['_add'])
            elif '_del' in v:
                ctx.owner._add_unlink_key(fname, v['_del'])



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_record import SimpleRecord
from tests.classes.typed_record import TypedRecord
from tests.classes.nullable_record import NullableRecord
from tests.classes.simple_weather import CamelizedWeather, UncamelizedWeather
from tests.classes.simple_quiz import SimpleQuiz


class TestDictOf(TestCase):

    def test_dictof_raises_if_value_is_not_dict(self):
        record = SimpleRecord(dict_record=5)
        self.assertRaises(ValidationException, record.validate)

    def test_dictof_raises_if_one_of_values_does_not_match_inner(self):
        record = SimpleRecord(dict_record={'a': '1', 'b': 2})
        self.assertRaises(ValidationException, record.validate)

    def test_dictof_does_not_raise_if_all_values_match_inner(self):
        record = SimpleRecord(dict_record={'a': '1', 'b': '2'})
        record.validate()

    def test_dictof_accepts_raw_type(self):
        record = SimpleRecord(dict_record={'a': '1', 'b': '2'})
        record.validate()
        record1 = SimpleRecord(dict_record={'a': '1', 'b': 2})
        self.assertRaises(ValidationException, record1.validate)

    def test_dictof_accepts_types_type(self):
        record = TypedRecord(dict_record={'a': '1', 'b': '2'})
        record.validate()
        record1 = TypedRecord(dict_record={'a': '1', 'b': 2})
        self.assertRaises(ValidationException, record1.validate)

    def test_dictof_does_not_allow_none_for_raw_typed_dict(self):
        record1 = SimpleRecord(dict_record={'a': '1', 'b': None})
        self.assertRaises(ValidationException, record1.validate)

    def test_dictof_does_not_allow_none_for_types_typed_dict(self):
        record1 = TypedRecord(dict_record={'a': '1', 'b': None})
        self.assertRaises(ValidationException, record1.validate)

    def test_dictof_allow_none_for_nullable_marked_typed_dict(self):
        record1 = NullableRecord(dict_record={'a': '1', 'b': None})
        record1.validate()

    def test_dictof_keep_keys_on_init(self):
        weather = CamelizedWeather(dataData={'lastDay': '2', 'nextDay': '4'})
        self.assertEqual(weather.data_data, {'lastDay': '2', 'nextDay': '4'})

    def test_dictof_keep_keys_on_tojson(self):
        weather = CamelizedWeather(dataData={'lastDay': '2', 'nextDay': '4'})
        self.assertEqual(weather.tojson()['dataData'],
                         {'lastDay': '2', 'nextDay': '4'})

    def test_dictof_validate_raises_for_one_item(self):
        quiz = SimpleQuiz(numbers={'a': 200,
                                   'b': 2,
                                   'c': 4,
                                   'd': 200,
                                   'e': 6,
                                   'f': 200})
        with self.assertRaises(ValidationException) as context:
            quiz.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['numbers.b'],
                         "value is not greater than or equal 100")



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_eq import SimpleEq
from jsonclasses.excs import ValidationException


class TestEq(TestCase):

    def test_eq_does_not_raise_if_equal(self):
        e = SimpleEq(eq_value="dsadsa")
        e.validate()

    def test_eq_raises_if_is_not_equal(self):
        e = SimpleEq(eq_value="dsa")
        with self.assertRaises(ValidationException) as context:
            e.validate()
        self.assertEqual(context.exception.keypath_messages['eqValue'],
                         "value is not equal")

    def test_eq_does_not_raise_if_equal_with_callable_value(self):
        e = SimpleEq(ceq_value="dsadsa")
        e.validate()

    def test_eq_raises_if_is_not_equal_with_callable_value(self):
        e = SimpleEq(ceq_value="dsa")
        with self.assertRaises(ValidationException) as context:
            e.validate()
        self.assertEqual(context.exception.keypath_messages['ceqValue'],
                         "value is not equal")

    def test_eq_does_not_raise_if_equal_with_types_value(self):
        e = SimpleEq(teq_value="dsadsa")
        e.validate()

    def test_eq_raises_if_is_not_equal_with_types_value(self):
        e = SimpleEq(teq_value="dsa")
        with self.assertRaises(ValidationException) as context:
            e.validate()
        self.assertEqual(context.exception.keypath_messages['teqValue'],
                         "value is not equal")



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleRecord:
    name: Optional[str]
    dict_record: dict[str, str] = types.nonnull.dictof(str)



from __future__ import annotations
from tests.classes.super_round import SuperRound
from unittest import TestCase



class TestCeil(TestCase):

    def test_ceil_transforms_float_value_into_int_by_ceiling(self):
        sr = SuperRound(ceil_value=19.8)
        self.assertEqual(sr.ceil_value, 20)

    def test_ceil_keeps_int_value_unchanged(self):
        sr = SuperRound(ceil_value=40)
        self.assertEqual(sr.ceil_value, 40)







from __future__ import annotations
from unittest import TestCase
from tests.classes.cellphone_description import CellphoneDescription


class TestToLower(TestCase):

    def test_tolower_lowercasefies_str(self):
        product = CellphoneDescription(cellphone_name='hello', cellphone_description='WORLD')
        self.assertEqual(product.cellphone_description, 'world')

    def test_tolower_lowercasefies_str_with_whitespace(self):
        product = CellphoneDescription(cellphone_name='hello', cellphone_description='WOR LD')
        self.assertEqual(product.cellphone_description, 'wor ld')


    def test_tolower_lowercasefies_str_with_int(self):
        product = CellphoneDescription(cellphone_name='hello', cellphone_description='WOR1LD')
        self.assertEqual(product.cellphone_description, 'wor1ld')

    def test_tolower_lowercasefies_str_with_special_characters(self):
        product = CellphoneDescription(cellphone_name='hello', cellphone_description='WOR@LD#')
        self.assertEqual(product.cellphone_description, 'wor@ld#')



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_secret import SimpleSecret


class TestTrim(TestCase):

    def test_trim_trims_str(self):
        secret = SimpleSecret(name='  1  ', message='2')
        self.assertEqual(secret.name, '1')



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .jobject import JObject


class VMsgCollector:

    def __init__(self: VMsgCollector) -> None:
        self.msgs: dict[str, str] = {}

    def receive(self: VMsgCollector, msgs: dict[str, str]) -> None:
        self.msgs.update(msgs)

    @property
    def messages(self: VMsgCollector) -> dict[str, str]:
        return self.msgs

    @property
    def has_msgs(self: VMsgCollector) -> bool:
        return len(self.msgs) > 0



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_inverse import SuperInverse


class TestInverse(TestCase):

    def test_inverse_transforms_false_if_bool_is_true(self):
        b = SuperInverse(iv=False)
        self.assertEqual(b.iv, True)

    def test_inverse_transforms_true_if_bool_is_false(self):
        b = SuperInverse(iv=True)
        self.assertEqual(b.iv, False)



"""This module defineds the JSON Class object mapping graph."""
from __future__ import annotations
from typing import Iterator, Union, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from .jobject import JObject


class MarkClassTable:

    def __init__(self):
        self._primary_key_table = {}
        self._memory_id_table = {}

    def put(self, object: JObject) -> None:
        pk = object._id
        if pk is None:
            memory_id = hex(id(object))
            self._memory_id_table[memory_id] = object
        else:
            self.putp(pk, object)

    def putp(self, pk: Union[str, int], object: JObject) -> None:
        """Put object to this class table when designated primary key. This is
        useful when transforming and the values are not set yet.
        """
        self._primary_key_table[str(pk)] = object

    def has(self, object: JObject) -> bool:
        pk = object._id
        if pk is not None:
            if self._primary_key_table.get(str(pk)) is not None:
                return True
        memory_id = hex(id(object))
        if self._memory_id_table.get(memory_id) is not None:
            return True
        return False

    def get(self, object: JObject) -> Optional[JObject]:
        pk = object._id
        if pk is not None:
            if self._primary_key_table.get(str(pk)) is not None:
                return self._primary_key_table.get(str(pk))
        memory_id = hex(id(object))
        if self._memory_id_table.get(memory_id) is not None:
            return self._memory_id_table.get(memory_id)
        return None

    def getp(self, pk: Union[str, int]) -> Optional[JObject]:
        return self._primary_key_table.get(str(pk))

    def getm(self, memid: int) -> Optional[JObject]:
        return self._memory_id_table.get(hex(memid))


class MGraph:
    """The mark graph is a graph containing JSON Class objects. It's used for
    marking objects as handled when performing validating and serializing.
    """

    def __init__(self):
        self._class_tables: dict[str, MarkClassTable] = {}

    def class_table(self, cls: type) -> MarkClassTable[JObject]:
        if self._class_tables.get(cls.__name__) is None:
            self._class_tables[cls.__name__] = MarkClassTable()
        return self._class_tables[cls.__name__]

    def put(self, object: JObject) -> None:
        self.class_table(object.__class__).put(object)

    def putp(self, pk: Union[str, int], object: JObject) -> None:
        self.class_table(object.__class__).putp(pk, object)

    def has(self, object: JObject) -> bool:
        return self.class_table(object.__class__).has(object)

    def get(self, object: JObject) -> JObject:
        return self.class_table(object.__class__).get(object)

    def getp(self, cls: type[JObject],
             pk: Union[str, int]) -> Optional[JObject]:
        return self.class_table(cls).getp(pk)

    def getm(self, cls: type[JObject],
             memid: int) -> Optional[JObject]:
        return self.class_table(cls).getm(memid)

    def __iter__(self) -> Iterator:
        lst = []
        for ct in self._class_tables.values():
            for obj in ct._primary_key_table.values():
                if obj not in lst:
                    lst.append(obj)
            for obj in ct._memory_id_table.values():
                if obj not in lst:
                    lst.append(obj)
        return lst.__iter__()



"""module for tonextday modifier."""
from __future__ import annotations
from datetime import datetime, date, timedelta
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToNextDayModifier(Modifier):
    """Change the day to the next day"""

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is datetime:
            return (ctx.val+timedelta(days=1)).replace(microsecond=0, second=0, minute=0, hour=0)
        if type(ctx.val) is date:
            return ctx.val+timedelta(days=1)
        else:
            return ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from classes.super_number import SuperNumber

class TestLt(TestCase):

    def test_lt_doesnt_raise_if_float_value_is_less_than_arg(self):
        n = SuperNumber(ff_lt=5.4)
        n.validate()

    def test_lt_raises_if_float_value_is_equal_to_arg(self):
        n = SuperNumber(ff_lt=5.5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ffLt'],
                         "value is not less than 5.5")

    def test_lt_raises_if_float_value_is_greater_than_arg(self):
        n = SuperNumber(ff_lt=5.6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ffLt'],
                         "value is not less than 5.5")

    def test_lt_doesnt_raise_if_int_value_is_less_than_arg(self):
        n = SuperNumber(if_lt=4)
        n.validate()

    def test_lt_raises_if_int_value_is_equal_to_arg(self):
        n = SuperNumber(if_lt=5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ifLt'],
                         "value is not less than 5")

    def test_lt_raises_if_int_value_is_greater_than_arg(self):
        n = SuperNumber(if_lt=6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['ifLt'],
                         "value is not less than 5")

    def test_lt_doesnt_raise_if_callable_value_is_less_than_arg(self):
        n = SuperNumber(c_lt=4)
        n.validate()

    def test_lt_raises_if_callable_value_is_equal_to_arg(self):
        n = SuperNumber(c_lt=5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['cLt'],
                         "value is not less than 5")

    def test_lt_raises_if_callable_value_is_greater_than_arg(self):
        n = SuperNumber(c_lt=6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['cLt'],
                         "value is not less than 5")

    def test_lt_doesnt_raise_if_types_value_is_less_than_arg(self):
        n = SuperNumber(t_lt=4)
        n.validate()

    def test_lt_raises_if_types_value_is_equal_to_arg(self):
        n = SuperNumber(t_lt=5)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['tLt'],
                         "value is not less than 5")

    def test_lt_raises_if_types_value_is_greater_than_arg(self):
        n = SuperNumber(t_lt=6)
        with self.assertRaises(ValidationException) as context:
            n.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['tLt'],
                         "value is not less than 5")



"""module for outputlname modifier."""
from ..fdef import FDef, EnumOutput
from .modifier import Modifier


class OutputLnameModifier(Modifier):
    """Output lname modifier tweaks enum modifier's behavior."""

    def define(self, fdef: FDef) -> None:
        fdef._enum_output = EnumOutput.LOWERCASE_NAME



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_bond import SuperBond


class TestUpperBond(TestCase):

    def test_upperbond_transforms_int_param_if_it_is_larger_than_int_val(self):
        n = SuperBond(i_ub=151)
        self.assertEqual(n.i_ub, 150)

    def test_upperbond_transforms_float_param_if_it_is_larger_than_float_val(self):
        n = SuperBond(f_ub=151.0)
        self.assertEqual(n.f_ub, 150.0)

    def test_upperbond_transforms_callable_param_if_it_is_larger_than_callable_val(self):
        n = SuperBond(c_ub=151)
        self.assertEqual(n.c_ub, 150)

    def test_upperbond_transforms_types_param_if_it_is_larger_than_types_val(self):
        n = SuperBond(t_ub=151)
        self.assertEqual(n.t_ub, 150)

    def test_upperbond_does_not_transform_int_param_if_it_is_less_than_int_val(self):
        n = SuperBond(i_ub=100)
        self.assertEqual(n.i_ub, 100)

    def test_upperbond_does_not_transform_float_param_if_it_is_less_than_float_val(self):
        n = SuperBond(f_ub=100.0)
        self.assertEqual(n.f_ub, 100.0)

    def test_upperbond_does_not_transform_callable_param_if_it_is_less_than_callable_val(self):
        n = SuperBond(c_ub=100)
        self.assertEqual(n.c_ub, 100)

    def test_upperbond_does_not_transform_types_param_if_it_is_less_than_types_val(self):
        n = SuperBond(t_ub=100)
        self.assertEqual(n.t_ub, 100)



from __future__ import annotations
from unittest import TestCase
from datetime import date, datetime, timezone
from jsonclasses.excs import ValidationException
from tests.classes.simple_balance import SimpleBalance


class TestDatetime(TestCase):

    def test_datetime_is_datetime_after_assigned(self):
        balance = SimpleBalance(date=datetime(2020, 11, 20, 0, 0, 1))
        self.assertEqual(balance._data_dict,
                         {'date': datetime(2020, 11, 20, 0, 0, 1),
                          'balance': None})

    def test_datetime_converts_date_into_datetime(self):
        balance = SimpleBalance(date=date(2020, 6, 30))
        self.assertEqual(balance.date, datetime(2020, 6, 30, 0, 0, 0, tzinfo=timezone.utc))

    def test_datetime_converts_date_str_into_datetime(self):
        balance = SimpleBalance(date='2020-11-20')
        self.assertEqual(balance.date, datetime(2020, 11, 20, 0, 0, 0, tzinfo=timezone.utc))
        self.assertEqual(type(balance.date), datetime)

    def test_datetime_converts_datetime_str_into_datetime(self):
        balance = SimpleBalance(date='2020-11-20T03:03:03.333Z')
        self.assertEqual(balance.date, datetime(2020, 11, 20, 3, 3, 3, 333000, tzinfo=timezone.utc))

    def test_datetime_raises_if_value_is_not_datetime(self):
        balance = SimpleBalance(date=True)
        with self.assertRaises(ValidationException) as context:
            balance.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['date'],
                         "value is not datetime")

    def test_datetime_is_datetime_str_when_tojson(self):
        balance = SimpleBalance(date='2020-11-20T03:03:03.333Z')
        self.assertEqual(balance.tojson(),
                         {'date': '2020-11-20T03:03:03.333Z'})



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperNumber:
    if_min: Optional[int] = types.int.min(5)
    if_max: Optional[int] = types.int.max(5)
    if_gte: Optional[int] = types.int.gte(5)
    if_lte: Optional[int] = types.int.lte(5)
    if_gt: Optional[int] = types.int.gt(5)
    if_lt: Optional[int] = types.int.lt(5)
    ff_min: Optional[float] = types.float.min(5.5)
    ff_max: Optional[float] = types.float.max(5.5)
    ff_gte: Optional[float] = types.float.gte(5.5)
    ff_lte: Optional[float] = types.float.lte(5.5)
    ff_gt: Optional[float] = types.float.gt(5.5)
    ff_lt: Optional[float] = types.float.lt(5.5)

    fcf_gt: Optional[float] = types.float.gt(lambda: 5.5)
    ftf_gt: Optional[float] = types.float.gt(types.default(5.5))

    c_lt: Optional[int] = types.int.lt(lambda: 5)
    t_lt: Optional[int] = types.int.lt(types.default(5))

    c_gte: Optional[int] = types.int.gte(lambda: 5)
    t_gte: Optional[int] = types.int.gte(types.default(5))

    c_lte: Optional[int] = types.int.lte(lambda: 5)
    t_lte: Optional[int] = types.int.lte(types.default(5))

    c_min: Optional[int] = types.int.min(lambda: 5)
    t_min: Optional[int] = types.int.min(types.default(5))

    c_max: Optional[int] = types.int.max(lambda: 5)
    t_max: Optional[int] = types.int.max(types.default(5))



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_article import SimpleArticle
from tests.classes.simple_language import SimpleLanguage
from tests.classes.simple_project import SimpleProject
from tests.classes.simple_chart import SimpleChart
from tests.classes.author import Author
from tests.classes.linked_profile import LinkedProfile
from tests.classes.linked_user import LinkedUser


class TestValidate(TestCase):

    def test_validate_does_not_raise_and_returns_self_for_valid_object(self):
        article = SimpleArticle(title='U Ia Huê', content='Bê Tshua Bo')
        self.assertEqual(article, article.validate())

    def test_validate_raises_if_object_is_not_valid(self):
        article = SimpleArticle()
        self.assertRaises(ValidationException, article.validate)

    def test_is_valid_returns_false_if_object_is_not_valid(self):
        article = SimpleArticle()
        self.assertEqual(False, article.is_valid)

    def test_is_valid_returns_true_if_object_is_valid(self):
        article = SimpleArticle(title='U Ia Huê', content='Bê Tshua Bo')
        self.assertEqual(True, article.is_valid)

    def test_validate_by_default_validates_one_field(self):
        article = SimpleArticle()
        with self.assertRaises(ValidationException) as context:
            article.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['title'],
                         "value required")

    def test_validate_validates_all_fields_if_option_is_passed(self):
        article = SimpleArticle()
        with self.assertRaises(ValidationException) as context:
            article.validate(all_fields=True)
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 2)
        self.assertEqual(exception.keypath_messages['title'],
                         "value required")
        self.assertEqual(exception.keypath_messages['content'],
                         "value required")

    def test_validate_validates_all_fields_if_class_config_is_on(self):
        language = SimpleLanguage()
        with self.assertRaises(ValidationException) as context:
            language.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 2)
        self.assertEqual(exception.keypath_messages['name'],
                         "value required")
        self.assertEqual(exception.keypath_messages['code'],
                         "value required")

    def test_validate_validates_one_field_if_explicitly_specified(self):
        language = SimpleLanguage()
        with self.assertRaises(ValidationException) as context:
            language.validate(all_fields=False)
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['name'],
                         "value required")

    def test_validate_validates_one_list_field_by_default(self):
        project = SimpleProject(name='Teo', attendees=['A', 'B', 'C', 'D'])
        with self.assertRaises(ValidationException) as context:
            project.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['attendees.0'],
                         "length of value is not greater than or equal 2")

    def test_validate_validates_all_list_fields_if_required(self):
        project = SimpleProject(name='Teo', attendees=['A', 'B', 'C', 'D'])
        with self.assertRaises(ValidationException) as context:
            project.validate(all_fields=True)
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 4)
        self.assertEqual(exception.keypath_messages['attendees.0'],
                         "length of value is not greater than or equal 2")
        self.assertEqual(exception.keypath_messages['attendees.1'],
                         "length of value is not greater than or equal 2")
        self.assertEqual(exception.keypath_messages['attendees.2'],
                         "length of value is not greater than or equal 2")
        self.assertEqual(exception.keypath_messages['attendees.3'],
                         "length of value is not greater than or equal 2")

    def test_validate_validates_one_dict_field_by_default(self):
        chart = SimpleChart(name='Teo', partitions={
            'a': 2, 'b': 3, 'c': 4, 'd': 5})
        with self.assertRaises(ValidationException) as context:
            chart.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['partitions.a'],
                         "value is not less than or equal 1")

    def test_validate_validates_all_dict_fields_if_required(self):
        chart = SimpleChart(name='Teo', partitions={
            'a': 2, 'b': 3, 'c': 4, 'd': 5})
        with self.assertRaises(ValidationException) as context:
            chart.validate(all_fields=True)
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 4)
        self.assertEqual(exception.keypath_messages['partitions.a'],
                         "value is not less than or equal 1")
        self.assertEqual(exception.keypath_messages['partitions.b'],
                         "value is not less than or equal 1")
        self.assertEqual(exception.keypath_messages['partitions.c'],
                         "value is not less than or equal 1")
        self.assertEqual(exception.keypath_messages['partitions.d'],
                         "value is not less than or equal 1")

    def test_validate_validates_with_class_config_by_default(self):
        author = Author(name='A', articles=[{}, {}])
        with self.assertRaises(ValidationException) as context:
            author.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 2)
        self.assertEqual(exception.keypath_messages['articles.0.title'],
                         "value required")
        self.assertEqual(exception.keypath_messages['articles.0.content'],
                         "value required")

    def test_validate_validates_one_field_inside_nested_if_required(self):
        author = Author(name='A', articles=[{}, {}])
        with self.assertRaises(ValidationException) as context:
            author.validate(all_fields=False)
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['articles.0.title'],
                         "value required")

    def test_validate_validates_all_field_inside_nested_if_required(self):
        author = Author(name='A', articles=[{}, {}])
        with self.assertRaises(ValidationException) as context:
            author.validate(all_fields=True)
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 4)
        self.assertEqual(exception.keypath_messages['articles.0.title'],
                         "value required")
        self.assertEqual(exception.keypath_messages['articles.0.content'],
                         "value required")
        self.assertEqual(exception.keypath_messages['articles.1.title'],
                         "value required")
        self.assertEqual(exception.keypath_messages['articles.1.content'],
                         "value required")

    def test_validate_only_validate_modified_fields_for_non_new_object(self):
        article = SimpleArticle(title='my', content='side')
        article._mark_not_new()
        article.content = None
        article._modified_fields = []
        article.validate(all_fields=True)

    def test_validate_validates_linked_objects_anyway(self):
        author = Author(name='Tsit Tiu',
                        articles=[{'title': 'Khua Tioh Sê Kai',
                                   'content': 'Ai Gua Tsuê'},
                                  {'title': 'Thên Ha Si Lan Ê',
                                   'content': 'Tsiu Ho Lang Tsai'}])
        author._mark_not_new()
        author.articles[0]._mark_not_new()
        author.articles[1]._mark_not_new()
        author.articles[0].content = None
        with self.assertRaises(ValidationException) as context:
            author.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['articles.0.content'],
                         "value required")

    def test_validate_linked_objects_no_infinite_loop(self):
        profile = LinkedProfile(name='Ua Bê Tshiu Pên')
        user = LinkedUser(name='Tsuan Sê Kai')
        user.profile = profile
        user.validate()



from __future__ import annotations
from typing import Any


def isjsonclass(cls: Any) -> bool:
    """Check if a class is jsonclass.

    Args:
        class_ (type): The class to check.

    Returns:
        bool: True if it's a jsonclass, otherwise False.
    """
    if not isinstance(cls, type):
        return False
    return hasattr(cls, '__is_jsonclass__')


def isjsonobject(object: Any) -> bool:
    """Check if an object is instance of jsonclass.

    Args:
        object (Any): The object to check.

    Returns:
        bool: True if it's an instance of jsonclass otherwise False.
    """
    return isjsonclass(object.__class__)



"""The owner observable dict."""
from __future__ import annotations
from typing import Generic, Any, Protocol, TypeVar, Union, MutableMapping
from collections.abc import Mapping, Iterable

KT_contra = TypeVar('KT_contra', contravariant=True)
VT_contra = TypeVar('VT_contra', contravariant=True)
_KT = TypeVar('_KT')
_VT = TypeVar('_VT')
_T = TypeVar('_T')


class DictOwner(Protocol[KT_contra, VT_contra]):

    def __odict_add__(self,
                      odict: OwnedDict,
                      key: KT_contra,
                      val: VT_contra) -> None: ...

    def __odict_del__(self, odict: OwnedDict, val: VT_contra) -> None: ...

    def __odict_will_change__(self, olist: OwnedDict) -> None: ...


def is_dict_owner(obj: Any):
    has_add = hasattr(obj, '__odict_add__')
    has_del = hasattr(obj, '__odict_del__')
    has_change = hasattr(obj, '__odict_will_change__')
    return has_add and has_del and has_change


class OwnedDict(dict, MutableMapping[_KT, _VT], Generic[_KT, _VT]):

    @property
    def owner(self) -> DictOwner:
        return self._owner

    @owner.setter
    def owner(self, val: DictOwner) -> None:
        self._owner = val

    @property
    def keypath(self) -> str:
        return self._keypath

    @keypath.setter
    def keypath(self, val: str) -> None:
        self._keypath = val

    def clear(self) -> None:
        values = [value for value in self.values()]
        self.owner.__odict_will_change__(self)
        super().clear()
        for value in values:
            self.owner.__odict_del__(self, value)

    def pop(self, *args) -> Union[_VT, _T]:
        len_args = len(args)
        if not (1 <= len_args <= 2):
            super().pop(*args)
        try:
            self.owner.__odict_will_change__(self)
            retval = super().pop(args[0])
            self.owner.__odict_del__(self, retval)
            return retval
        except KeyError as e:
            if len_args > 1:
                return args[1]
            raise e

    def popitem(self) -> tuple[_KT, _VT]:
        self.owner.__odict_will_change__(self)
        result = super().popitem()
        self.owner.__odict_del__(self, result[1])
        return result

    def setdefault(self, *args) -> _VT:
        will_set = False
        len_args = len(args)
        if len_args == 2:
            try:
                self[args[0]]
            except KeyError:
                will_set = True
        if will_set:
            self.owner.__odict_will_change__(self)
        retval = super().setdefault(*args)
        if will_set:
            self.owner.__odict_add__(self, args[0], args[1])
        return retval

    def update(self, *args, **kwargs: _VT) -> None:
        len_args = len(args)
        if not (0 <= len_args <= 1):
            return super().update(*args, **kwargs)
        if len_args == 0:
            new_dict = kwargs
        elif isinstance(args[0], Mapping):
            new_dict = dict(args[0], **kwargs)
        elif isinstance(args[0], Iterable):
            new_dict = dict(args[0], **kwargs)
        else:
            new_dict = {}
        items_to_del = [i[1] for i in self.items() if i[0] in new_dict.keys()]
        self.owner.__odict_will_change__(self)
        super().update(*args, **kwargs)
        for v in items_to_del:
            self.owner.__odict_del__(self, v)
        for k, v in new_dict.items():
            self.owner.__odict_add__(self, k, v)

    def __setitem__(self, k: _KT, v: _VT) -> None:
        remove_callback = False
        removed = None
        try:
            removed = self[k]
            remove_callback = True
        except KeyError:
            pass
        self.owner.__odict_will_change__(self)
        super().__setitem__(k, v)
        if remove_callback:
            self.owner.__odict_del__(self, removed)
        self.owner.__odict_add__(self, k, v)

    def __delitem__(self, k: _KT) -> None:
        try:
            removed = self[k]
        except KeyError:
            removed = None
        self.owner.__odict_will_change__(self)
        super().__delitem__(k)
        self.owner.__odict_del__(self, removed)

    def __ior__(self, rhs: Iterable) -> OwnedDict[_KT, _VT]:
        new_dict = dict(rhs)
        items_to_del = [i[1] for i in self.items() if i[0] in new_dict.keys()]
        self.owner.__odict_will_change__(self)
        retval = super().__ior__(rhs)
        for v in items_to_del:
            self.owner.__odict_del__(self, v)
        for k, v in new_dict.items():
            self.owner.__odict_add__(self, k, v)
        return retval



"""module for padstart modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class PadStartModifier(Modifier):
    """PadStart modifier add str to the start of str value."""

    def __init__(self, char: str | Callable | Types, target_length: int | Callable | Types) -> None:
        self.char = char
        self.target_length = target_length

    def transform(self, ctx: Ctx) -> Any:
        char = self.resolve_param(self.char, ctx)
        target_length = self.resolve_param(self.target_length, ctx)
        if type(ctx.val) is str:
            pad_str = (target_length - len(ctx.val)) * char
            return pad_str + ctx.val
        else:
            return ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleUserAccount:
    email: Optional[str] = types.str.presentwithout('phone_number')
    phone_number: Optional[str] = types.str.presentwithout('email')



from __future__ import annotations
from unittest import TestCase
from tests.classes.super_filter import SuperFilter


class TestFilter(TestCase):

    def test_filter_filters_list_by_callable(self):
        item = SuperFilter(list1=[1, 2, 3, 4])
        self.assertEqual(item.list1, [2, 4])

    def test_filter_filters_list_by_types(self):
        item = SuperFilter(list2=[1, 2, 3, 4])
        self.assertEqual(item.list2, [2, 4])



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_secret import SimpleSecret


class TestInternal(TestCase):

    def test_internal_cannot_be_set_through_init(self):
        secret = SimpleSecret(name='1', message='2')
        self.assertEqual(secret._data_dict, {'name': '1', 'message': None})

    def test_internal_cannot_be_set_through_set(self):
        secret = SimpleSecret(name='1', message='2')
        secret.set(message='4')
        self.assertEqual(secret._data_dict, {'name': '1', 'message': None})

    def test_internal_can_be_set_through_update(self):
        secret = SimpleSecret(name='1', message='2')
        secret.update(message='4')
        self.assertEqual(secret._data_dict, {'name': '1', 'message': '4'})

    def test_internal_can_be_set_through_assign(self):
        secret = SimpleSecret(name='1', message='2')
        secret.message = '5'
        self.assertEqual(secret._data_dict, {'name': '1', 'message': '5'})

    def test_internal_wont_appear_in_output(self):
        secret = SimpleSecret(name='1', message='2')
        secret.message = '5'
        self.assertEqual(secret.tojson(), {'name': '1'})



"""module for numeric modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class NumericModifier(Modifier):
    """Numeric modifier raises if value is not a numeric."""

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, str) and not ctx.val.isnumeric():
            ctx.raise_vexc('value is not numeric string')




from __future__ import annotations
from unittest import TestCase
from tests.classes.cellphone_name import CellphoneName


class TestToCap(TestCase):

    def test_tocap_capitalize_simple_str(self):
        product = CellphoneName(cellphone_name='hello', cellphone_title='world')
        self.assertEqual(product.cellphone_name, 'Hello')

    def test_tocap_capitalize_str_contains_whitespace(self):
        product = CellphoneName(cellphone_name='hello world', cellphone_title='world')
        self.assertEqual(product.cellphone_name, 'Hello world')

    def test_tocap_capitalize_str_contains_int(self):
        product = CellphoneName(cellphone_name='hell1o world', cellphone_title='world')
        self.assertEqual(product.cellphone_name, 'Hell1o world')

    def test_tocap_capitalize_str_contains_special_character(self):
        product = CellphoneName(cellphone_name='hel!lo world', cellphone_title='world')
        self.assertEqual(product.cellphone_name, 'Hel!lo world')


"""module for uploader modifier."""
from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
from ..fdef import FDef, FSubtype
from ..uploaders import request_uploader, S3Uploader, AliOSSUploader
if TYPE_CHECKING:
    from ..ctx import Ctx


class UploaderModifier(Modifier):
    """Upload file stream to cloud storage and get the string url back."""

    def __init__(self, arg: str | Callable) -> None:
        self.arg = arg
        if callable(arg):
            params_len = len(signature(arg).parameters)
            if params_len > 2 or params_len < 1:
                raise ValueError('not a valid transformer')
        else:
            self.check_packages()

    def define(self, fdef: FDef) -> None:
        fdef._fsubtype = FSubtype.FILE

    def packages(self) -> dict[str, (str, str)] | None:
        if type(self.arg) is str:
            uploader = request_uploader(self.arg)
            if isinstance(uploader, S3Uploader):
                return {'boto3': ('boto3', '>=1.18.61,<2.0.0')}
            elif isinstance(uploader, AliOSSUploader):
                return {'oss2': ('oss2', '>=2.0.0,<3.0.0')}
            else:
                return {}
        return {}

    def transform(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if callable(self.arg):
            params_len = len(signature(self.arg).parameters)
            if params_len == 1:
                return self.arg(ctx.val)
            elif params_len == 2:
                return self.arg(ctx.val, ctx)
        else:
            uploader = request_uploader(self.arg)
            return uploader.upload(ctx.val)



from __future__ import annotations
from tests.classes.simple_calculation import SimpleCalculation
from unittest import TestCase

class TestMod(TestCase):


    def test_mod_with_int_value_mods_original_value(self):
        mod_int = SimpleCalculation(i_mod=8)
        self.assertEqual(mod_int.i_mod, 3)

    def test_mod_with_float_value_mods_original_value(self):
        mod_float = SimpleCalculation(f_mod=8.5)
        self.assertEqual(mod_float.f_mod, 1)

    def test_mod_mods_callable_value_to_original_value(self):
        mod_float = SimpleCalculation(c_mod=8.4)
        self.assertEqual(mod_float.c_mod, 1.4000000000000004)

    def test_mod_mods_types_value_to_original_value(self):
        mod_float = SimpleCalculation(t_mod=8.4)
        self.assertEqual(mod_float.t_mod, 1.4000000000000004)



from __future__ import annotations
from random import random
from tests.classes.super_random import SuperRandom
from unittest import TestCase

class TestRandomFloat(TestCase):

    def test_random_float_generate_a_random_float_value(self):
        i = SuperRandom(random_float=10.0)
        self.assertEqual(i.random_float, 10.0)

    def test_random_float_generate_a_random_float_value_with_callable(self):
        i = SuperRandom(random_cfloat=10.0)
        self.assertEqual(i.random_cfloat, 10.0)

    def test_random_float_generate_a_random_float_value_with_types(self):
        i = SuperRandom(random_tfloat=10.0)
        self.assertEqual(i.random_tfloat, 10.0)



"""module for reset modifier."""
from .modifier import Modifier
from ..fdef import FDef


class ResetModifier(Modifier):
    """A reset modifier marks fields for recording value before modified.
    This is used for comparing and validating values on update.
    """

    def define(self, fdef: FDef) -> None:
        fdef._has_reset_modifier = True



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleWordbook:
    name: Optional[str]
    words: list[str] = types.nonnull.listof(str)



from datetime import date, datetime
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperDate:
    dbd: Optional[date] = types.date.before(date(2020, 4, 3))
    dbdt: Optional[date] = types.date.before(datetime(2020, 4, 3, 0, 0))
    dtbdt: Optional[datetime] = types.datetime.before(datetime(2020, 4, 3, 0, 0))
    dtbd: Optional[datetime] = types.datetime.before(date(2020, 4, 3))

    dad: Optional[date] = types.date.after(date(2020, 4, 3))
    dadt: Optional[date] = types.date.after(datetime(2020, 4, 3, 0, 0))
    dtadt: Optional[datetime] = types.datetime.after(datetime(2020, 4, 3, 0, 0))
    dtad: Optional[datetime] = types.datetime.after(date(2020, 4, 3))

    dbcd: Optional[date] = types.date.before(lambda: date(2020, 4, 3))
    dbtd: Optional[date] = types.date.before(types.default(date(2020, 4, 3)))

    dacd: Optional[date] = types.date.after(lambda: date(2020, 4, 3))
    datd: Optional[date] = types.date.after(types.default(date(2020, 4, 3)))



"""This module defines all excs that JSON classes uses."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from inspect import getmodule, getsourcelines
if TYPE_CHECKING:
    from .jobject import JObject


class UnresolvedTypeNameException(Exception):
    """Unresolved type name in graph.
    """

    def __init__(self, msg: str) -> None:
        self.msg = msg
        super().__init__(msg)


class UnlinkableJSONClassException(Exception):
    """This exception is raised when jsonclass objects are linked and no unique
    primary key value is found.
    """

    def __init__(self, class_: type, has_field: bool) -> None:
        """Create an exception that represents an object is not valid for
        putting into object graph, thus cannot be linked.

        Args:
            has_field (bool): Whether this class defined primary field.
            class_ (type): The jsonclass class.
        """
        self.has_field = has_field
        self.class_ = class_
        class_name = class_.__name__
        if not has_field:
            self.message = f"please define a primary field on `{class_name}`"
        else:
            self.message = (f"primary key value not found on `{class_name}` "
                            "object")
        super().__init__(self.message)


class JSONClassRedefinitionException(Exception):
    """This exception is raised when user defines a JSON class with a name that
    exists before in the same graph.
    """

    def __init__(self, new_class: type, exist_class: type[JObject]) -> None:
        """Create an exception that notifies the user that a class with
        duplicated name is defined twice.

        Args:
            new_class (type): The new class which is putting on the graph.
            exist_class(type): The existing class which the user defined.
        """
        name = new_class.__name__
        original_module = getmodule(exist_class)
        assert original_module is not None
        original_file = original_module.__file__
        original_line = getsourcelines(exist_class)[1]
        new_module = getmodule(new_class)
        assert new_module is not None
        new_file = new_module.__file__
        new_line = getsourcelines(new_class)[1]
        graph = exist_class.cdef.jconf.cgraph
        message = (f'jsonclass name conflict in graph `{graph}`: '
                   f'exist `{name}` defined at '
                   f'`{original_file}:{original_line}`, '
                   f'new `{name}` defined at `{new_file}:{new_line}`')
        super().__init__(message)


class JSONClassTypedDictRedefinitionException(Exception):
    """This exception is raised when user defines a typed dict with a name that
    exists before in the same graph.
    """

    def __init__(self, new_class: type, exist_class: type[JObject]) -> None:
        """Create an exception that notifies the user that a typed dict class
        with duplicated name is defined twice.

        Args:
            new_class (type): The new class which is putting on the graph.
            exist_class(type): The existing class which the user defined.
        """
        name = new_class.__name__
        original_module = getmodule(exist_class)
        assert original_module is not None
        original_file = original_module.__file__
        original_line = getsourcelines(exist_class)[1]
        new_module = getmodule(new_class)
        assert new_module is not None
        new_file = new_module.__file__
        new_line = getsourcelines(new_class)[1]
        graph = exist_class.cdef.jconf.cgraph
        message = (f'jsonclass typed dict name conflict in graph `{graph}`: '
                   f'exist `{name}` defined at '
                   f'`{original_file}:{original_line}`, '
                   f'new `{name}` defined at `{new_file}:{new_line}`')
        super().__init__(message)


class JSONClassNotFoundException(Exception):
    """This exception is raised when a JSON class with name is not found on a
    graph.
    """
    def __init__(self, class_name: str, graph_name: str):
        message = (f'JSON Class with name \'{class_name}\' in graph '
                   f'\'{graph_name}\' is not found.')
        super().__init__(message)


class JSONClassTypedDictNotFoundException(Exception):
    """This exception is raised when a typed dict with name is not found on a
    graph.
    """
    def __init__(self, class_name: str, graph_name: str):
        message = (f'jsonclass typed dict with name \'{class_name}\' in graph '
                   f'\'{graph_name}\' is not found.')
        super().__init__(message)


class LinkedFieldUnmatchException(Exception):
    """This exception is raised when a JSON class reference field doesn't match
    the counterpart on the other side.
    """
    def __init__(self, class1: str, field1: str, class2: str, field2: str):
        message = (f"unmatched linked field {class1}.{field1} and "
                   f"{class2}.{field2}")
        super().__init__(message)


class ObjectNotFoundException(Exception):
    """ObjectNotFoundException is designed to be raised by jsonclasses ORM
    integration implementations. Server authors and jsonclasses server
    integration authors should catch this to present error to frontend clients.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class UniqueConstraintException(Exception):
    """UniqueConstraintException is designed to be raised by JSON Classes ORM
    integration implementations. When saving objects into the database, if
    object violates the uniqueness rule, this exception should be raised.
    """

    def __init__(self, field: str | list[str], message: str | None = None):
        self.message = message or (f'value is not unique')
        if type(field) is str:
            self.keypath_messages = {
                field: self.message
            }
        else:
            self.keypath_messages = {k: self.message for k in field}
        super().__init__(self.message)


class ValidationException(Exception):
    """ValidationException is throwed on jsonclass object validation or on
    eager validation. Server authors and jsonclasses server integration authors
    should catch this to present error to frontend clients.
    """

    def __init__(self, keypath_messages: dict[str, str], root: Any):
        self.keypath_messages = keypath_messages
        self.message = self.formatted_keypath_messages()
        self.root = root
        super().__init__(self.message)

    def formatted_keypath_messages(self):
        """The formatted keypath message for print."""
        retval = 'Json classes validation failed:\n'
        for k, v in self.keypath_messages.items():
            retval += f'  \'{k}\': {v}\n'
        return retval


class UnauthorizedActionException(Exception):
    """UnauthorizedActionException is raised when operator checking guards are
    not passed.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class DeletionDeniedException(Exception):
    """DeletionDeniedException is raised on jsonclass object deletion.
    """

    def __init__(self) -> None:
        self.message = 'deletion denied'
        super().__init__(self.message)


class AbstractJSONClassException(Exception):
    """Abstract class should not be initialized nor serialized into database.
    When an abstract JSON class is initialized, this error should be raised.
    """

    def __init__(self, class_: type) -> None:
        self.class_ = class_
        self.message = (f'{class_.__name__} is an abstract class and should '
                        'not be initialized')
        super().__init__(self.message)


class JSONClassResetError(Exception):
    """This error is raised when an ORM object is new and thus cannot be reset.
    """

    def __init__(self) -> None:
        self.message = 'object is new and cannot be reset'
        super().__init__(self.message)


class JSONClassResetNotEnabledError(Exception):
    """This error is raised when calling reset on an object which class doesn't
    enable `reset_all_fields`.
    """

    def __init__(self) -> None:
        self.message = 'reset called on a reset disabled object'
        super().__init__(self.message)


class JSONClassGraphMergeConflictException(Exception):
    """This exception is raised when there is an object conflict when linking
    jsonclass objects.
    """

    def __init__(self, message: str, object: Any) -> None:
        super().__init__(f'{message}{object}')



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperInt:
    i_o: Optional[int] = types.int.odd
    i_e: Optional[int] = types.int.even
    i_a: Optional[int] = types.int.abs



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class RoleUser:
    username: str = types.str.required
    role: str = types.str.readonly.default('normal').required



"""module for onupdate modifier."""
from __future__ import annotations
from typing import Callable, Any, cast, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class OnUpdateModifier(Modifier):
    """On update modifier is called when value is modified and saving is
    triggered.
    """

    def __init__(self, callback: Callable) -> None:
        if not callable(callback):
            raise ValueError('onupdate argument is not callable')
        params_len = len(signature(callback).parameters)
        if params_len > 3:
            raise ValueError('not a valid onupdate callable')
        self.callback = callback

    def serialize(self, ctx: Ctx) -> Any:
        from ..jobject import JObject
        name = ctx.keypathp[-1]
        parent = cast(JObject, ctx.parent)
        if name not in parent.previous_values:
            return ctx.val
        prev_value = parent.previous_values[name]
        params_len = len(signature(self.callback).parameters)
        if params_len == 0:
            self.callback()
        elif params_len == 1:
            self.callback(ctx.val)
        elif params_len == 2:
            self.callback(prev_value, ctx.val)
        elif params_len == 3:
            self.callback(prev_value,
                          ctx.val,
                          ctx)
        return ctx.val



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperBond:
    i_ub: Optional[int] = types.int.upperbond(150)
    f_ub: Optional[float] = types.float.upperbond(150.0)
    c_ub: Optional[int] = types.int.upperbond(lambda: 150)
    t_ub: Optional[int] = types.int.upperbond(types.default(150))

    i_lb: Optional[int] = types.int.lowerbond(1)
    f_lb: Optional[float] = types.float.lowerbond(1.0)
    c_lb: Optional[int] = types.int.lowerbond(lambda: 1)
    t_lb: Optional[int] = types.float.lowerbond(types.default(1))



from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkedthru, types
if TYPE_CHECKING:
    from tests.classes.linked_default_user import LinkedDefaultUser


@jsonclass
class LinkedDefaultPost:
    name: str = types.str.default('Untitled')
    users: Annotated[list[LinkedDefaultUser], linkedthru('posts')]



from __future__ import annotations
from datetime import date, datetime
from tests.classes.super_date import SuperDate
from unittest import TestCase
from jsonclasses.excs import ValidationException

class TestAfter(TestCase):

    def test_after_dosent_raise_if_date_value_is_after_date_point(self):
        dad = SuperDate(dad=date(2020, 4, 4))
        dad.validate()

    def test_after_dosent_raise_if_date_value_is_after_datetime_piont(self):
        dadt = SuperDate(dadt=date(2020, 4, 4))
        dadt.validate()

    def test_after_dosent_raise_if_datetime_value_is_after_date_piont(self):
        dtad = SuperDate(dtad=datetime(2020, 4, 4, 12, 30))
        dtad.validate()

    def test_after_raise_if_date_value_and_date_point_are_the_same(self):
        dad = SuperDate(dad=date(2020, 4, 3))
        with self.assertRaises(ValidationException) as context:
            dad.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dad'],
                         "value is too early")

    def test_after_raise_if_date_value_and_datetime_point_are_the_same(self):
        dadt = SuperDate(dadt=date(2020, 4, 3))
        with self.assertRaises(ValidationException) as context:
            dadt.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dadt'],
                         "value is too early")

    def test_after_raise_if_datetime_value_and_datetime_point_are_the_same(self):
        dtad = SuperDate(dtad=datetime(2020, 4, 3, 0, 0))
        with self.assertRaises(ValidationException) as context:
            dtad.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dtad'],
                         "value is too early")


    def test_after_raise_if_date_value_earlier_than_date_point(self):
        dad = SuperDate(dad=date(2020, 4, 2))
        with self.assertRaises(ValidationException) as context:
            dad.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dad'],
                         "value is too early")

    def test_after_raise_if_date_value_earlier_than_datetime_point(self):
        dadt = SuperDate(dadt=date(2020, 4, 2))
        with self.assertRaises(ValidationException) as context:
            dadt.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dadt'],
                         "value is too early")

    def test_after_raise_if_datetime_value_earlier_than_date_point(self):
        dtad = SuperDate(dtad=datetime(2020, 4, 2, 0, 0))
        with self.assertRaises(ValidationException) as context:
            dtad.validate()
        self.assertEqual(len(context.exception.keypath_messages), 1)
        self.assertEqual(context.exception.keypath_messages['dtad'],
                         "value is too early")

    def test_after_does_not_raise_if_date_value_is_after_callable_date_point(self):
        dacd = SuperDate(dacd=date(2020, 4, 4))
        dacd.validate()

    def test_after_does_not_raise_if_date_value_is_after_types_date_piont(self):
        datd = SuperDate(datd=date(2020, 4, 4))
        datd.validate()



"""module for isthis modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class IsThisModifier(Modifier):
    """Is this modifier tests value against the context owner."""


    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            ctx.raise_vexc('none is not this')
        if ctx.val is ctx.owner:
            return None
        if ctx.val.__class__ == ctx.owner.__class__:
            if ctx.val._id == ctx.owner._id:
                return None
        ctx.raise_vexc('value is not this')



from __future__ import annotations
from unittest import TestCase
from tests.classes.local_key_author import LKAuthor, LKArticle
from tests.classes.linked_author import LinkedAuthor
from tests.classes.linked_article import LinkedArticle


class TestLocalKey(TestCase):

    def test_local_key_can_be_accessed(self):
        article = LKArticle(name='A')
        self.assertEqual(article.author_id, None)

    def test_set_object_modifies_local_key(self):
        author = LKAuthor(name='A')
        article = LKArticle(name='A', author=author)
        self.assertEqual(article.author_id, author.id)

    def test_set_local_key_resets_object(self):
        author = LKAuthor(name='A')
        article = LKArticle(name='A', author=author)
        self.assertEqual(article.author, author)
        self.assertEqual(author.articles[0], article)
        article.author_id = 500
        self.assertEqual(article.author, None)
        self.assertEqual(author.articles, [])

    def test_local_key_is_set_to_none_then_object_is_set_to_none(self):
        author = LKAuthor(name='A')
        article = LKArticle(name='A', author=author)
        self.assertEqual(article.author, author)
        self.assertEqual(author.articles[0], article)
        article.author_id = None
        self.assertEqual(article.author, None)
        self.assertEqual(author.articles, [])

    def test_object_is_set_to_none_then_local_key_is_set_to_none(self):
        author = LKAuthor(name='A')
        article = LKArticle(name='A', author=author)
        self.assertEqual(article.author, author)
        self.assertEqual(author.articles[0], article)
        article.author = None
        self.assertEqual(article.author_id, None)
        self.assertEqual(author.articles, [])

    def test_object_is_set_to_none_o_then_local_key_is_set_to_none(self):
        author = LKAuthor(name='A')
        article = LKArticle(name='A', author=author)
        self.assertEqual(article.author, author)
        self.assertEqual(author.articles[0], article)
        author.articles = []
        self.assertEqual(article.author_id, None)
        self.assertEqual(author.articles, [])

    def test_local_key_is_considered_modified_field(self):
        author = LKAuthor(name='A')
        article = LKArticle(name='A', author=author)
        setattr(author, '_is_new', False)
        setattr(article, '_is_new', False)
        article.author = None
        self.assertEqual(article.modified_fields, ('author',))
        self.assertEqual(author.modified_fields, ('articles',))

    def test_local_key_is_considered_modified_field_with_id_none_assign(self):
        author = LKAuthor(name='A')
        article = LKArticle(name='A', author=author)
        setattr(author, '_is_new', False)
        setattr(article, '_is_new', False)
        article.author_id = None
        self.assertEqual(article.modified_fields, ('author',))
        self.assertEqual(author.modified_fields, ('articles',))

    def test_local_key_is_considered_modified_field_with_id_assign(self):
        article = LKArticle(name='A', author=None)
        author = LKAuthor(name='A')
        setattr(author, '_is_new', False)
        setattr(article, '_is_new', False)
        article.author_id = author.id
        self.assertEqual(article.modified_fields, ('author',))

    def test_local_key_is_considered_modified_field_even_id_is_none(self):
        article = LinkedArticle(name='A', author=None)
        author = LinkedAuthor(name='A')
        setattr(author, '_is_new', False)
        setattr(article, '_is_new', False)
        article.author = author
        self.assertEqual(article.author_id, None)
        self.assertEqual(article.modified_fields, ('author',))
        self.assertEqual(author.modified_fields, ('articles',))

    def test_local_key_set_on_save_should_not_reset_value(self):
        article = LinkedArticle(name='A', author=None)
        article.author_id = 995
        article._set_on_save()
        self.assertEqual(article.author_id, 995)

    def test_init_should_accept_local_key(self):
        article = LKArticle(name='A', author_id=1005)
        self.assertEqual(article.author, None)
        self.assertEqual(article.author_id, 1005)

    def test_set_should_accept_local_key(self):
        article = LKArticle()
        article.set(name='A', author_id=1005)
        self.assertEqual(article.author, None)
        self.assertEqual(article.author_id, 1005)

    def test_update_should_accept_local_key(self):
        article = LKArticle()
        article.update(name='A', author_id=1005)
        self.assertEqual(article.author, None)
        self.assertEqual(article.author_id, 1005)

    def test_init_should_accept_local_key_in_json_form(self):
        article = LKArticle(**{'name': 'A', 'authorId': 1005})
        self.assertEqual(article.author, None)
        self.assertEqual(article.author_id, 1005)

    def test_set_should_accept_local_key_in_json_form(self):
        article = LKArticle()
        article.set(**{'name': 'A', 'authorId': 1005})
        self.assertEqual(article.author, None)
        self.assertEqual(article.author_id, 1005)



"""module for match modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from re import compile, match, IGNORECASE
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


# https://stackoverflow.com/questions/7160737/how-to-validate-a-url-in-python-malformed-or-not
url_regex = compile(
    r'^(?:http|ftp)s?://'  # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
    r'localhost|'  # localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
    r'(?::\d+)?'  # optional port
    r'(?:/?|[/?]\S+)$', IGNORECASE)

class UrlModifier(Modifier):
    """URL modifier raises if value is not valid url."""

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, str) and match(url_regex, ctx.val) is None:
            ctx.raise_vexc('value is not valid url string')



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperSuffix:
    s_hs: Optional[str] = types.str.hassuffix('python')
    l_hs: Optional[list[str]] = types.listof(str).hassuffix(['dd', 'ee', 'ff'])
    c_hs: Optional[str] = types.str.hassuffix(lambda: 'python')
    c_l_hs: Optional[list[str]] = types.listof(str).hassuffix(lambda: ['dd', 'ee', 'ff'])
    t_hs: Optional[str] = types.str.hassuffix(types.default('python'))
    t_l_hs: Optional[list[str]] = types.listof(str).hassuffix(types.default(['dd', 'ee', 'ff']))

    s_iso: Optional[str] = types.str.issuffixof('python')
    l_iso: Optional[list[str]] = types.listof(str).issuffixof(['ee', 'ff'])
    c_iso: Optional[str] = types.str.issuffixof(lambda: 'python')
    c_l_iso: Optional[list[str]] = types.listof(str).issuffixof(lambda: ['ee', 'ff'])
    t_iso: Optional[str] = types.str.issuffixof(types.default('python'))
    t_l_iso: Optional[list[str]] = types.listof(str).issuffixof(types.default(['ee', 'ff']))



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class EvUser:
    username: str = types.str.required
    password: str = types.str.minlength(8).maxlength(16) \
                         .transform(lambda s: s + '0x0x').required


@jsonclass
class EvUserL:
    passwords: list[str] = types.listof(
        types.str.minlength(2).maxlength(4)
                 .transform(lambda s: s + '0x0x0x0x'))


@jsonclass
class EvUserD:
    passwords: dict[str, str] = types.dictof(
        types.str.minlength(2).maxlength(4)
                 .transform(lambda s: s + '0x0x0x0x'))



from __future__ import annotations
from jsonclasses import jsonclass, types


def check_owner(product: GMProduct, operator: GMProductUser) -> bool:
    return product.user.id == operator.id


def check_tier(product: GMProduct, operator: GMProductUser) -> bool:
    return operator.paid_user


@jsonclass
class GMProductUser:
    id: str
    name: str
    paid_user: bool
    books: list[GMProduct] = types.nonnull.listof('GMProduct').linkedby('user')


@jsonclass(can_delete=[check_owner, check_tier], can_read=[check_owner, check_tier])
class GMProduct:
    name: str
    user: GMProductUser = types.objof('GMProductUser').linkto.required



"""module for tolower modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ToLowerModifier(Modifier):
    """convert string into lower case"""

    def transform(self, ctx: Ctx) -> Any:
        return ctx.val.lower() if isinstance(ctx.val, str) else ctx.val



"""module for join modifier."""
from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types

class JoinModifier(Modifier):
    """Join modifier Joins a list of string values into a string."""

    def __init__(self, sep: str | Callable | Types) -> None:
        self.sep = sep

    def transform(self, ctx: Ctx) -> Any:
        return self.resolve_param(self.sep, ctx).join(ctx.val) if isinstance(ctx.val, list) else ctx.val



"""module for hexcolor modifier."""
from __future__ import annotations
from typing import Any, cast, TYPE_CHECKING
from re import compile, match
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


hex_color_regex = compile('[0-9a-fA-F]{6}')

class HexColorModifier(Modifier):
    """Email modifier raises if value is not valid email."""

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is str:
            strval = cast(str, ctx.val)
            if strval.startswith('#'):
                return strval[1:]
        return ctx.val

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, str) and match(hex_color_regex, ctx.val) is None:
            ctx.raise_vexc('value is not hex color string')


"""module for insert at modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx
    from ..types import Types


class InsertAtModifier(Modifier):
    """Insert at modifier for inserting an element to the list"""

    def __init__(self, item: Any | Callable | Types, index: int | Callable | Types) -> None:
        self.item = item
        self.index = index

    def transform(self, ctx: Ctx) -> Any:
        index = self.resolve_param(self.index, ctx)
        item = self.resolve_param(self.item, ctx)
        if isinstance(ctx.val, list):
            return ctx.val[:index] + [item] + ctx.val[index:]
        elif type(ctx.val) is str:
            return ctx.val[:index] + str(item) + ctx.val[index:]
        else:
            return ctx.val



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_article import SimpleArticle
from tests.classes.linked_profile import LinkedProfile
from tests.classes.linked_user import LinkedUser


class TestRequired(TestCase):

    def test_required_raises_when_embedded_value_is_none(self):
        article = SimpleArticle()
        self.assertRaisesRegex(
            ValidationException,
            "'title': value required",
            article.validate)

    def test_required_raises_when_local_key_is_none(self):
        profile = LinkedProfile(name='profile')
        self.assertRaisesRegex(
            ValidationException,
            "'user': value required",
            profile.validate)

    def test_required_does_not_raise_when_local_key_is_present(self):
        profile = LinkedProfile(name='profile')
        setattr(profile, 'user_id', 5)
        profile.validate()

    def test_required_does_not_raise_when_local_key_value_is_present(self):
        profile = LinkedProfile(name='profile', user=LinkedUser(name='ok'))
        profile.validate()



"""module for outputname modifier."""
from ..fdef import FDef, EnumOutput
from .modifier import Modifier


class OutputNameModifier(Modifier):
    """Output name modifier tweaks enum modifier's behavior."""

    def define(self, fdef: FDef) -> None:
        fdef._enum_output = EnumOutput.NAME



from __future__ import annotations
from tests.classes.simple_calculation import SimpleCalculation
from unittest import TestCase

class TestAdd(TestCase):


    def test_add_adds_constant_int_value_to_original_value(self):
        add_int = SimpleCalculation(i_add=8)
        self.assertEqual(add_int.i_add, 13)

    def test_add_adds_constant_float_value_to_original_value(self):
        add_float = SimpleCalculation(f_add=8.5)
        self.assertEqual(add_float.f_add, 11)

    def test_add_adds_callable_value_to_original_value(self):
        add_float = SimpleCalculation(c_add=8.4)
        self.assertEqual(add_float.c_add, 10.9)

    def test_add_adds_types_value_to_original_value(self):
        add_float = SimpleCalculation(t_add=8.4)
        self.assertEqual(add_float.t_add, 10.9)



from __future__ import annotations
from datetime import date, datetime
from unittest import TestCase

from tests.classes.super_datetime import SuperDateTime


class TestToboyear(TestCase):

    def test_toboyear_transforms_datetime_into_the_time_of_beginning_of_year(self):
        d = SuperDateTime(dtby=datetime(2021, 10, 11, 17, 37, 27,43235))
        self.assertEqual(d.dtby, datetime(2021,1, 1, 0, 0))

    def test_toboyear_transforms_date_into_the_time_of_beginning_of_year(self):
        d = SuperDateTime(dby=date(2021, 10, 11))
        self.assertEqual(d.dby, date(2021,1, 1))

    def test_toboyear_does_not_transform_if_is_not_datetime(self):
        s = SuperDateTime(sby="12345")
        self.assertEqual(s.sby, "12345")



"""module for required modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class PresentWithModifier(Modifier):
    """Fields marked with presentwith modifier are forced presented if
    referring field is present. If referring field has None value, this field's
    value is optional. If referring field has non None value, value of this
    field is required.
    """

    def __init__(self, referring_key: str) -> None:
        self.referring_key = referring_key

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is not None:
            return
        try:
            referred_value = getattr(ctx.owner, self.referring_key)
        except AttributeError:
            raise ValueError(f'unexist referring key \'{self.referring_key}\' '
                             'passed to present with modifier')
        if referred_value is not None and ctx.val is None:
            ctx.raise_vexc(f"value is not present with '{self.referring_key}'")



"""module for linkedby modifier."""
from ..fdef import DeleteRule, FDef, FStore, FType, Nullability
from .modifier import Modifier


class LinkedByModifier(Modifier):
    """Linked by modifier marks a field linked with a foreign key."""

    def __init__(self, foreign_key: str) -> None:
        self.foreign_key = foreign_key

    def define(self, fdef: FDef) -> None:
        fdef._fstore = FStore.FOREIGN_KEY
        fdef._foreign_key = self.foreign_key
        fdef._use_join_table = False
        if fdef._delete_rule is None:
            fdef._delete_rule = DeleteRule.NULLIFY
        if fdef._ftype == FType.LIST:
            if fdef._collection_nullability is Nullability.UNDEFINED:
                fdef._collection_nullability = Nullability.NONNULL



from __future__ import annotations
from datetime import datetime
from unittest import TestCase
from jsonclasses.excs import JSONClassGraphMergeConflictException
from tests.classes.blog import User, Post


class TestGraph(TestCase):

    def test_graph_is_assigned_on_create(self):
        user = User(id=1, name='Phuê Ê')
        self.assertEqual(user._graph.get(user), user)
        objects = [object for object in user._graph]
        self.assertEqual(objects, [user])

    def test_graph_is_merged_on_connect(self):
        user = User(id=1, name='Phuê Ê')
        post = Post(id=1, name='M Tsai Tai Kha Kui Ê Lang')
        user.posts.append(post)
        self.assertEqual(user._graph, post._graph)
        self.assertEqual(user._graph.get(user), user)
        self.assertEqual(user._graph.get(post), post)
        objects = [object for object in user._graph]
        self.assertEqual(objects, [post, user])

    def test_graph_is_merged_on_init(self):
        post = Post(id=1, name='Tsih Ma Khai Si',
                    updated_at=datetime(2021, 3, 10, 0, 0, 0))
        user = User(id=1, name='Tsu Iu Tsu Tsai', posts=[post])
        self.assertEqual(user._graph, post._graph)
        self.assertEqual(user._graph.get(user), user)
        self.assertEqual(user._graph.get(post), post)
        objects = [object for object in user._graph]
        self.assertEqual(objects, [user, post])

    def test_graph_on_conflict_raises(self):
        post = Post(id=1, name='M Tsai Tai Kha Kui Ê Lang',
                    updated_at=datetime(2021, 3, 10, 0, 0, 0))
        user = User(id=1, name='Phuê Ê')
        user.posts = [post]
        user._mark_not_new()
        post._mark_not_new()
        post_new = Post(id=1, name='Koh Kin Tiu Koh Tsin Ki Thai',
                        updated_at=datetime(2021, 3, 11, 0, 0, 0))
        post_new._mark_not_new()
        with self.assertRaises(JSONClassGraphMergeConflictException):
            user.posts = [post_new]



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class ValidPassword:
    name: Optional[str]
    password: str = types.str.validate(lambda p: len(p) % 2 == 0).required


@jsonclass
class ValidPasswordMessage:
    name: Optional[str]
    password: str = types.str.validate(lambda p: None if len(p) % 2 == 0
                                       else 'wrong').required


@jsonclass
class CValidPassword:
    name: Optional[str]
    password: str = types.str.validate(lambda p, c: c.val == p).required


@jsonclass
class TValidPassword:
    name: Optional[str]
    password: int = types.int.validate(types.add(5).eq(50)).required


@jsonclass
class OptionalPassword:
    name: Optional[str]
    password: Optional[str] = types.str.validate(lambda p: len(p) % 2 == 0)



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SuperPermissionUser:
    id: int = types.int.primary.required
    code: int = types.int.required
    articles: list[SuperPermissionArticle] = types.listof('SuperPermissionArticle').linkedby('user')


@jsonclass
class SuperPermissionArticle:
    id: int = types.int.primary.required
    code: int = types.int.required
    user: SuperPermissionUser = types.objof('SuperPermissionUser').linkto
    title: str = types.str.canc(types.getop.fval('code').eq(types.this.fval('code'))).required
    content: str = types.str.canr(types.getop.fval('code').eq(types.this.fval('code'))).required
    content2: Optional[str] = types.str.canu(types.getop.isobj(types.this.fval('user')))



"""This module defines `JConf`. Each JSON class has its own configuration. The
configuration object tweaks the behavior of JSON classes.
"""
from __future__ import annotations
from typing import Callable, Literal, Any, cast, final, TYPE_CHECKING
from .jobject import JObject
from .keypath import camelize_key, underscore_key, identical_key
if TYPE_CHECKING:
    from .jfield import JField
    from .cgraph import CGraph
    from .types import Types


OnCreate = Callable[[JObject, Any], None]
OnUpdate = Callable[[JObject, Any], None]
OnDelete = Callable[[JObject, Any], None]
CanCreate = Callable[[JObject, Any], bool | None | str]
CanUpdate = Callable[[JObject, Any], bool | None | str]
CanDelete = Callable[[JObject, Any], bool | None | str]
CanRead = Callable[[JObject, Any], bool | None | str]


@final
class JConf:
    """The configuration of JSON classes. Each JSON class has its own
    configuration that each instance shares. This object tweaks the behavior of
    JSON classes.
    """

    def __init__(self: JConf,
                 cgraph: str | None,
                 output_key_strategy: Callable[[str], str] | Literal['identical', 'camelize'] | None,
                 input_key_strategy: Callable[[str], str] | Literal['identical', 'underscore'] | None,
                 strict_input: bool | None,
                 ref_name_strategy: Callable[[JField], str] | None,
                 validate_all_fields: bool | None,
                 abstract: bool | None,
                 reset_all_fields: bool | None,
                 output_null: bool | None,
                 on_create: OnCreate | list[OnCreate] | Types | None,
                 on_update: OnUpdate | list[OnUpdate] | Types | None,
                 on_delete: OnDelete | list[OnDelete] | Types | None,
                 can_create: CanCreate | list[CanCreate] | Types | None,
                 can_update: CanUpdate | list[CanUpdate] | Types | None,
                 can_delete: CanDelete | list[CanDelete] | Types | None,
                 can_read: CanRead | list[CanRead] | Types | None) -> None:
        """
        Initialize a new configuration object.

        Args:
            cgraph (Optional[str]): The name of the class graph on which \
                the JSON class is defined.
            output_key_strategy (Optional[Callable[[str], str]]): How \
                object keys are encoded. The default is camelize.
            input_key_strategy (Optional[Callable[[str], str]]): How \
                object keys are decoded. The default is underscore.
            strict_input (Optional[bool]): Whether raise errors on receiving \
                invalid input keys.
            ref_name_strategy (Optional[Callable[[JField], str]]): The \
                reference field local key conversion function.
            validate_all_fields (Optional[bool]): The default field \
                validating method when performing saving and validating.
            abstract (Optional[bool]): Instance of abstract classes cannot \
                be initialized.
            reset_all_fields (Optional[bool]): Whether record all previous \
                values of an object and enable reset functionality.
            output_null: (Optional[bool]): Whether output null on tojson
                instead of leaving unexisting fields.
            on_create (Optional[Union[OnCreate, list[OnCreate]]]): The callback
                on first time save.
            on_update (Optional[Union[OnUpdate, list[OnUpdate]]]): The callback
                on existing object save.
            on_delete (Optional[Union[OnDelete, list[OnDelete]]]): The callback
                on existing object deletion.
            can_create (Optional[Union[CanCreate, list[CanCreate]]]): The
                creation guard.
            can_update (Optional[Union[CanUpdate, list[CanUpdate]]]): The
                updation guard.
            can_delete (Optional[Union[CanDelete, list[CanDelete]]]): The
                deletion guard.
            can_read (Optional[Union[CanRead, list[CanRead]]]): The reading
                guard.
        """
        from .types import Types
        self._cls: type[JObject] | None = None
        self._cgraph = cgraph or 'default'
        self._output_key_strategy = _figure_out(output_key_strategy)
        self._input_key_strategy = _figure_out(input_key_strategy)
        self._strict_input = strict_input
        self._ref_name_strategy = ref_name_strategy
        self._validate_all_fields = validate_all_fields
        self._abstract = abstract
        self._reset_all_fields = reset_all_fields
        self._output_null = output_null
        if callable(on_create) or isinstance(on_create, Types):
            self._on_create = [on_create]
        elif isinstance(on_create, list):
            self._on_create = on_create
        else:
            self._on_create = []
        if callable(on_update) or isinstance(on_update, Types):
            self._on_update = [on_update]
        elif isinstance(on_update, list):
            self._on_update = on_update
        else:
            self._on_update = []
        if callable(on_delete) or isinstance(on_delete, Types):
            self._on_delete = [on_delete]
        elif isinstance(on_delete, list):
            self._on_delete = on_delete
        else:
            self._on_delete = []
        if callable(can_create) or isinstance(can_create, Types):
            self._can_create = [can_create]
        elif isinstance(can_create, list):
            self._can_create = can_create
        else:
            self._can_create = []
        if callable(can_update) or isinstance(can_update, Types):
            self._can_update = [can_update]
        elif isinstance(can_update, list):
            self._can_update = can_update
        else:
            self._can_update = []
        if callable(can_delete) or isinstance(can_delete, Types):
            self._can_delete = [can_delete]
        elif isinstance(can_delete, list):
            self._can_delete = can_delete
        else:
            self._can_delete = []
        if callable(can_read) or isinstance(can_read, Types):
            self._can_read = [can_read]
        elif isinstance(can_read, list):
            self._can_read = can_read
        else:
            self._can_read = []

    def __eq__(self: JConf, other: Any) -> bool:
        if not isinstance(other, JConf):
            return False
        other_config = cast(JConf, other)
        if self.cgraph != other_config.cgraph:
            return False
        if self.output_key_strategy != other_config.output_key_strategy:
            return False
        if self.input_key_strategy != other_config.input_key_strategy:
            return False
        if self.strict_input != other_config.strict_input:
            return False
        if self.ref_name_strategy != other_config.ref_name_strategy:
            return False
        if self.validate_all_fields != other_config.validate_all_fields:
            return False
        if self.abstract != other_config.abstract:
            return False
        if self.reset_all_fields != other_config.reset_all_fields:
            return False
        if self.output_null != other_config.output_null:
            return False
        if self.on_create != other_config.on_create:
            return False
        if self.on_update != other_config.on_update:
            return False
        if self.on_delete != other_config._on_delete:
            return False
        if self.can_create != other_config.can_create:
            return False
        if self.can_update != other_config.can_update:
            return False
        if self.can_delete != other_config.can_delete:
            return False
        if self.can_read != other_config.can_read:
            return False
        return True

    @property
    def cls(self: JConf) -> type[JObject]:
        """The JSON class on which this class config is defined.
        """
        return cast(type[JObject], self._cls)

    @property
    def cgraph(self: JConf) -> CGraph:
        """The name of the class graph on which the JSON class is defined.
        """
        from .cgraph import CGraph
        return CGraph(self._cgraph)

    @property
    def output_key_strategy(self: JConf) -> Callable[[str], str]:
        """The object key encoding strategy.
        """
        if self._output_key_strategy is None:
            return self.cgraph.default_config.output_key_strategy
        return self._output_key_strategy

    @property
    def input_key_strategy(self: JConf) -> Callable[[str], str]:
        """The object key decoding strategy.
        """
        if self._input_key_strategy is None:
            return self.cgraph.default_config.input_key_strategy
        return self._input_key_strategy

    @property
    def strict_input(self: JConf) -> bool:
        """Whether raise errors on receiving invalid input keys.
        """
        if self._strict_input is None:
            return self.cgraph.default_config.strict_input
        return self._strict_input

    @property
    def ref_name_strategy(self: JConf) -> Callable[[JField], str]:
        """The reference field local key conversion function.
        """
        if self._ref_name_strategy is None:
            return self.cgraph.default_config.ref_name_strategy
        return self._ref_name_strategy

    @property
    def validate_all_fields(self: JConf) -> bool:
        """The default field validating method when performing saving and
        validating.
        """
        if self._validate_all_fields is None:
            return self.cgraph.default_config.validate_all_fields
        return self._validate_all_fields

    @property
    def abstract(self: JConf) -> bool:
        """Instance of abstract classes cannot be initialized.
        """
        if self._abstract is None:
            return self.cgraph.default_config.abstract
        return self._abstract

    @property
    def reset_all_fields(self: JConf) -> bool:
        """Whether record all previous values of an object and enable reset
        functionality.
        """
        if self._reset_all_fields is None:
            return self.cgraph.default_config.reset_all_fields
        return self._reset_all_fields

    @property
    def output_null(self: JConf) -> bool:
        """Whether output null values in the result json.
        """
        if self._output_null is None:
            return self.cgraph.default_config.output_null
        return self._output_null

    @property
    def on_create(self: JConf) -> list[OnCreate | Types]:
        """The object creation callback.
        """
        return self._on_create

    @property
    def on_update(self: JConf) -> list[OnUpdate | Types]:
        """The object saving callback.
        """
        return self._on_update

    @property
    def on_delete(self: JConf) -> list[OnDelete | Types]:
        """The object deleting callback.
        """
        return self._on_delete

    @property
    def can_create(self: JConf) -> list[CanCreate | Types]:
        """The object creation guard.
        """
        return self._can_create

    @property
    def can_update(self: JConf) -> list[CanUpdate | Types]:
        """The object updation guard.
        """
        return self._can_update

    @property
    def can_delete(self: JConf) -> list[CanDelete | Types]:
        """The object deletion guard.
        """
        return self._can_delete

    @property
    def can_read(self: JConf) -> list[CanRead | Types]:
        """The object reading guard.
        """
        return self._can_read


def _figure_out(s: Callable[[str], str] | Literal['identical', 'underscore', 'camelize'] | None) -> Callable[[str], str]:
    if s is None:
        return None
    if type(s) is str:
        match s:
            case 'underscore':
                return underscore_key
            case 'camelize':
                return camelize_key
            case _:
                return identical_key
    return s



from __future__ import annotations
from jsonclasses import jsonclass


def increase_updated_count(book: CBOBook, operator: int) -> None:
    book.updated_count += operator


@jsonclass(on_update=increase_updated_count)
class CBOBook:
    name: str
    content: str
    updated_count: int = 0



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.super_message import SuperMessage


class TestVMsgModifier(TestCase):

    def test_vmsg_outputs_custom_error_message_on_invalid(self):
        message = SuperMessage(name='123')
        with self.assertRaises(ValidationException) as context:
            message.validate()
        self.assertEqual(context.exception.keypath_messages, {
            'name': 'MUST 666'
        })

    def test_vmsg_does_not_output_custom_error_message_on_valid(self):
        message = SuperMessage(name='123456')
        message.validate()



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class CalcUser:
    name: str
    first_name: str = types.str.getter(lambda u: u.name.split(" ")[0])
    last_name: str = types.str.getter(types.this.fval('name').split(' ').at(1))
    base_score: float
    score: float = types.float.getter(types.this.fval('base_score').mul(2)).negative



def assignfn(fn: str, u: SetterUser):
    ln = u.last_name
    u.name = fn + " " + ln


def assignln(ln: str, u: SetterUser):
    fn = u.first_name
    u.name = fn + " " + ln


@jsonclass
class SetterUser:
    name: str
    first_name: str = types.str.getter(lambda u: u.name.split(" ")[0]).setter(assignfn)
    last_name: str = types.str.getter(types.this.fval('name').split(' ').at(1)).setter(assignln)
    base_score: float
    score: float = types.float.getter(types.this.fval('base_score').mul(2)).setter(types.this.assign('base_score', types.passin.div(2)))



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_order import SimpleOrder
from tests.classes.simple_project import SimpleProject
from tests.classes.simple_chart import SimpleChart
from tests.classes.author import Author
from tests.classes.nested_object import NestedDict, NestedList
from tests.classes.linked_song import LinkedSong, LinkedSinger


class TestIsModified(TestCase):

    def test_jobject_is_not_modified_by_default(self):
        order = SimpleOrder(quantity=5)
        order._mark_not_new()
        self.assertEqual(order.is_modified, False)

    def test_jobject_is_modified_is_readonly(self):
        order = SimpleOrder(quantity=5)
        with self.assertRaises(AttributeError) as context:
            order.is_modified = True
        self.assertEqual(str(context.exception), "can't set attribute")

    def test_jobject_is_modified_if_any_field_is_modified(self):
        order = SimpleOrder(quantity=5)
        order._mark_not_new()
        order.quantity = 2
        self.assertEqual(order.modified_fields, ('quantity',))
        self.assertEqual(order.is_modified, True)

    def test_jobject_is_modified_if_list_value_is_mutated(self):
        project = SimpleProject(name='OK', attendees=['Butai', 'Buti'])
        project._mark_not_new()
        project.attendees.append('Liittiengesai')
        self.assertEqual(project.modified_fields, ('attendees',))
        self.assertEqual(project.is_modified, True)

    def test_jobject_is_modified_if_dict_value_is_mutated(self):
        chart = SimpleChart(name='B', partitions={'thenha': 0.5, 'guaê': 0.3})
        chart._mark_not_new()
        chart.partitions['lanê'] = 0.8
        self.assertEqual(chart.modified_fields, ('partitions',))
        self.assertEqual(chart.is_modified, True)

    def test_jobject_is_not_modified_if_nested_is_mutated(self):
        author = Author(name='Kia',
                        articles=[{'title': 'Long',
                                   'content': 'Khng Ti Ua E Sim Lai'}])
        author._mark_not_new()
        author.articles[0]._mark_not_new()
        author.articles[0].title = 'M Guan Kong Tsai Huê'
        self.assertEqual(author.is_modified, False)
        self.assertEqual(author.modified_fields, ())
        self.assertEqual(author.articles[0].is_modified, True)
        self.assertEqual(author.articles[0].modified_fields, ('title',))

    def test_jobject_is_modified_if_nested_list_is_modified(self):
        nlst = NestedList(id='1', value={'a': ['1', '2'], 'b': ['1', '2']})
        nlst._mark_not_new()
        nlst.value['b'].append('3')
        self.assertEqual(nlst.is_modified, True)
        self.assertEqual(nlst.modified_fields, ('value.b',))

    def test_jobject_is_modified_if_nested_dict_is_modified(self):
        ndct = NestedDict(id='1', value=[{'a': '1'}, {'b': '1'}])
        ndct._mark_not_new()
        ndct.value[1]['c'] = '2'
        self.assertEqual(ndct.is_modified, True)
        self.assertEqual(ndct.modified_fields, ('value.1',))

    def test_jobject_is_modified_if_local_key_list_is_assigned(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singer_ids = [3, 4]
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))

    def test_jobject_is_modified_if_local_key_list_is_assigned_nothing_changed_to_object_list(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singer_ids = [1, 2, 3]
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))


    def test_jobject_is_modified_if_local_key_list_is_appended(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singer_ids.append(3)
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))

    def test_jobject_is_modified_if_local_key_list_is_truncated(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singer_ids.remove(2)
        self.assertEqual(song.singers, [singer1])
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))

    def test_jobject_is_modified_if_local_key_list_is_truncated_nothing_changed_to_object_list(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song.singer_ids.append(5)
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singer_ids.remove(5)
        self.assertEqual(song.singers, [singer1, singer2])
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))


    def test_jobject_is_modified_if_local_key_field_list_is_assigned(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        singer3 = LinkedSinger(id=3)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singers = [singer1, singer2, singer3]
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))

    def test_jobject_is_not_modified_if_local_key_field_list_is_assigned_nothing_changed_at_all(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singers = [singer1, singer2]
        self.assertEqual(song.is_modified, False)
        self.assertEqual(song.modified_fields, ())

    def test_jobject_is_modified_if_local_key_field_list_is_appended(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        singer3 = LinkedSinger(id=3)
        song = LinkedSong(id=1, singers=[singer1, singer2])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singers.append(singer3)
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))

    def test_jobject_is_modified_if_local_key_field_list_is_truncated(self):
        singer1 = LinkedSinger(id=1)
        singer2 = LinkedSinger(id=2)
        singer3 = LinkedSinger(id=3)
        song = LinkedSong(id=1, singers=[singer1, singer2, singer3])
        song._mark_not_new()
        self.assertEqual(song.is_modified, False)
        song.singers.remove(singer3)
        self.assertEqual(song.is_modified, True)
        self.assertEqual(song.modified_fields, ('singers',))



"""This module contains utility methods for working with owned collections.
"""
from __future__ import annotations
from typing import TypeVar, Any
from .odict import OwnedDict, DictOwner
from .olist import OwnedList, ListOwner
from .keypath import concat_keypath

KT = TypeVar('KT')
VT = TypeVar('VT')
T = TypeVar('T')


def to_owned_dict(owner: DictOwner,
                  dct: dict[KT, VT],
                  keypath: str) -> OwnedDict[KT, VT]:
    new_dct = {}
    for k, v in dct.items():
        if isinstance(v, list):
            new_dct[k] = to_owned_list(owner, v, concat_keypath(keypath, k))
        elif isinstance(v, dict):
            new_dct[k] = to_owned_dict(owner, v, concat_keypath(keypath, k))
        else:
            new_dct[k] = v
    owned_dict = OwnedDict[Any](new_dct)
    owned_dict.keypath = keypath
    owned_dict.owner = owner
    return owned_dict


def to_owned_list(owner: ListOwner,
                  lst: list[T],
                  keypath: str) -> OwnedList[T]:
    new_lst = []
    for i, v in enumerate(lst):
        if isinstance(v, list):
            new_lst.append(to_owned_list(owner, v, concat_keypath(keypath, i)))
        elif isinstance(v, dict):
            new_lst.append(to_owned_dict(owner, v, concat_keypath(keypath, i)))
        else:
            new_lst.append(v)
    owned_list = OwnedList[Any](new_lst)
    owned_list.keypath = keypath
    owned_list.owner = owner
    return owned_list


def unowned_copy_dict(any_dict: dict[KT, VT]) -> dict[KT, VT]:
    retval = {}
    for k, v in any_dict.items():
        if isinstance(v, dict):
            retval[k] = unowned_copy_dict(v)
        elif isinstance(v, list):
            retval[k] = unowned_copy_list(v)
        else:
            retval[k] = v
    return retval


def unowned_copy_list(any_list: list[T]) -> list[T]:
    retval = []
    for v in any_list:
        if isinstance(v, dict):
            retval.append(unowned_copy_dict(v))
        elif isinstance(v, list):
            retval.append(unowned_copy_list(v))
        else:
            retval.append(v)
    return retval



"""module for deny modifier."""
from ..fdef import FDef, DeleteRule
from .modifier import Modifier


class DenyModifier(Modifier):
    """Deny modifier marks a relationship's delete rule as deny."""

    def define(self, fdef: FDef) -> None:
        fdef._delete_rule = DeleteRule.DENY



"""module for positive modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class PositiveModifier(Modifier):
    """Positive modifier marks value valid for large than zero."""

    def validate(self, ctx: Ctx) -> None:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        if is_number and ctx.val <= 0:
            ctx.raise_vexc('value is not positive')



"""module for tonextmonth modifier."""
from __future__ import annotations
from datetime import datetime, date
from typing import Any, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx

class ToNextMonModifier(Modifier):
    """Change the month to the next month"""

    def transform(self, ctx: Ctx) -> Any:
        if type(ctx.val) is datetime:
            year, month = divmod(ctx.val.month + 1, 12)
            if month == 0:
                month = 12
                year = year -1
            next_month = datetime(ctx.val.year + year, month, 1, tzinfo=ctx.val.tzinfo)
            return next_month
        if type(ctx.val) is date:
            year, month = divmod(ctx.val.month + 1, 12)
            if month == 0:
                month = 12
                year = year -1
            next_month = date(ctx.val.year + year, month, 1)
            return next_month
        else:
            return ctx.val



from __future__ import annotations
from typing import Annotated, TYPE_CHECKING
from jsonclasses import jsonclass, linkedby
if TYPE_CHECKING:
    from tests.classes.linked_article import LinkedArticle


@jsonclass
class LinkedAuthor:
    name: str
    articles: Annotated[list[LinkedArticle], linkedby('author')]



from __future__ import annotations
from unittest import TestCase
from tests.classes.simple_wordbook import SimpleWordbook
from tests.classes.linked_school import LinkedSchool
from tests.classes.linked_student import LinkedStudent
from tests.classes.simple_record import SimpleRecord


class TestNonnull(TestCase):

    def test_nonnull_primitive_list_has_default_value_empty_list(self):
        book = SimpleWordbook(name='Wordbook A')
        self.assertEqual(book.words, [])
        book1 = SimpleWordbook(name='WOrdbook B', words=['tsa', 'mbo'])
        self.assertEqual(book1.words, ['tsa', 'mbo'])

    def test_nonnull_linked_list_has_default_value_empty_list(self):
        school = LinkedSchool(name='School')
        self.assertEqual(school.students, [])
        student = LinkedStudent(name='Student 1')
        school1 = LinkedSchool(name='School 1', students=[student])
        self.assertEqual(school1.students, [student])

    def test_nonnull_dict_has_default_value_empty_dict(self):
        record = SimpleRecord(name='R')
        self.assertEqual(record.dict_record, {})
        record1 = SimpleRecord(name='R', dict_record={'a': '1'})
        self.assertEqual(record1.dict_record, {'a': '1'})
